import typing
import base64
import secrets
import asyncio

from . import model  as _model
from . import enums  as _enums
from . import events as _events
from . import client as _client


__all__ = ('get_image_data_uri', 'get_eventful_intents', 'interact')


def _get_image_data_uri(mime, data):

    data = base64.b64encode(data)

    return f'data:{mime};base64:{data}'


def get_image_date_uri(data: bytes, 
                       mime: str = 'image/png'):

    """
    Get an :ddoc:`image date uri </reference#image-data>`.

    :param data:
        The data to encode within the uri.
    :param mime:
        The data mimetype to attach. 
    """

    return _get_image_data_uri(mime, data)


_eventful_intents = {
    _events.CreateAutoModerationRule          : _enums.Intents.guild_moderation,
    _events.UpdateAutoModerationRule          : _enums.Intents.auto_moderation_configuration,
    _events.DeleteAutoModerationRule          : _enums.Intents.auto_moderation_configuration,
    _events.ExecuteAutoModerationRule         : _enums.Intents.auto_moderation_action_execution,
    _events.CreateChannel                     : _enums.Intents.guilds,
    _events.UpdateChannel                     : _enums.Intents.guilds,
    _events.DeleteChannel                     : _enums.Intents.guilds,
    _events.CreateThread                      : _enums.Intents.guilds,
    _events.UpdateThread                      : _enums.Intents.guilds,
    _events.DeleteThread                      : _enums.Intents.guilds,
    _events.SyncThreads                       : _enums.Intents.guilds,
    _events.UpdateThreadMember                : _enums.Intents.guilds,
    _events.UpdateThreadMembers               : _enums.Intents.guilds,
    _events.UpdateChannelPins                 : _enums.Intents.guilds | _enums.Intents.direct_messages,
    _events.AvailableGuild                    : _enums.Intents.guilds,
    _events.UpdateChannel                     : _enums.Intents.guilds,
    _events.CreateGuild                       : _enums.Intents.guilds,
    _events.DeleteGuild                       : _enums.Intents.guilds,
    _events.UnavailableGuild                  : _enums.Intents.guilds,
    _events.UpdateGuild                       : _enums.Intents.guilds,
    _events.CreateGuildAuditLogEntry          : _enums.Intents.guild_moderation,
    _events.CreateGuildBan                    : _enums.Intents.guild_moderation,
    _events.DeleteGuildBan                    : _enums.Intents.guild_moderation,
    _events.UpdateGuildEmojis                 : _enums.Intents.guild_emojis_and_stickers,
    _events.UpdateGuildStickers               : _enums.Intents.guild_emojis_and_stickers,
    _events.UpdateGuildIntegrations           : _enums.Intents.guild_integrations,
    _events.CreateGuildMember                 : _enums.Intents.guild_members,
    _events.DeleteGuildMember                 : _enums.Intents.guild_members,
    _events.UpdateGuildMember                 : _enums.Intents.guild_members,
    _events.CreateGuildRole                   : _enums.Intents.guilds,
    _events.UpdateGuildRole                   : _enums.Intents.guilds,
    _events.DeleteGuildRole                   : _enums.Intents.guilds,
    _events.CreateGuildScheduledEvent         : _enums.Intents.guild_scheduled_events,
    _events.UpdateGuildScheduledEvent         : _enums.Intents.guild_scheduled_events,
    _events.DeleteGuildScheduledEvent         : _enums.Intents.guild_scheduled_events,
    _events.CreateGuildScheduledEventUser     : _enums.Intents.guild_scheduled_events,
    _events.DeleteGuildScheduledEventUser     : _enums.Intents.guild_scheduled_events,
    _events.CreateIntegration                 : _enums.Intents.guild_integrations,
    _events.UpdateIntegration                 : _enums.Intents.guild_integrations,
    _events.DeleteIntegration                 : _enums.Intents.guild_integrations,
    _events.CreateInvite                      : _enums.Intents.guild_invites,
    _events.DeleteInvite                      : _enums.Intents.guild_invites,
    _events.CreateMessage                     : _enums.Intents.guild_messages | _enums.Intents.direct_messages | _enums.Intents.message_content,
    _events.UpdateMessage                     : _enums.Intents.guild_messages | _enums.Intents.direct_messages | _enums.Intents.message_content,
    _events.DeleteMessage                     : _enums.Intents.guild_messages | _enums.Intents.direct_messages,
    _events.DeleteMessages                    : _enums.Intents.guild_messages,
    _events.CreateMessageReaction             : _enums.Intents.guild_message_reactions | _enums.Intents.direct_message_reactions,
    _events.DeleteMessageReaction             : _enums.Intents.guild_message_reactions | _enums.Intents.direct_message_reactions,
    _events.DeleteAllMessageReactions         : _enums.Intents.guild_message_reactions | _enums.Intents.direct_message_reactions,
    _events.DeleteAllMessageEmojiReactions    : _enums.Intents.guild_message_reactions | _enums.Intents.direct_message_reactions,
    _events.UpdatePresence                    : _enums.Intents.guild_presences,
    _events.CreateTypingIndicator             : _enums.Intents.guild_message_typing,
    _events.UpdateVoiceState                  : _enums.Intents.guild_voice_states, # shall always be included
    _events.UpdateWebhooks                    : _enums.Intents.guild_webhooks,
    _events.CreateStageInstance               : _enums.Intents.guilds,
    _events.UpdateStageInstance               : _enums.Intents.guilds,
    _events.DeleteStageInstance               : _enums.Intents.guilds,
}


def _get_eventful_intents(Events: list[typing.Any]):

    intents = _enums.Intents(0)

    for Event in Events:
        try:
            subinents = _eventful_intents[Event]
        except KeyError:
            continue
        intents |= subinents

    return intents


def get_eventful_intents(Events):

    """
    Get the intents required to receive the specified events.
    """

    return _get_eventful_intents(Events)


_interact_types = {
    _model.enums.InteractionType.message_component, 
    _model.enums.InteractionResponseType.modal
}


def _interact(client, callback, component, / , **kwargs):

    custom_id = secrets.token_hex(32)

    component['custom_id'] = custom_id

    async def check(core_event):
        return core_event.interaction.type in _interact_types and core_event.interaction.data.custom_id == custom_id
    
    sentinel = client.wait(_events.CreateInteraction, check, **kwargs)

    async def sub_callback(core_event):
        response = await callback(core_event)
        if not response is None:
            await client.create_interaction_response(core_event.interaction.id, core_event.interaction.token, **response)

    def done_callback(task):
        try:
            core_event = task.result()
        except TimeoutError:
            core_event = None
        asyncio.create_task(sub_callback(core_event))

    sentinel.add_done_callback(done_callback)

    return component


def interact(client   : _client.Client, 
             callback : typing.Callable[[typing.Union[typing.Any, None]], typing.Awaitable[_model.protocols.InteractionResponse | None]], 
             component: _model.protocols.MessageButtonComponent | _model.protocols.MessageSelectMenuComponent | _model.protocols.MessageTextInputComponent,
             /,
             timeout  : float = 3):
    
    """
    Invoke a callback when the component has been interacted with.

    If a :code:`.model.protocols.InteractionResponse` is returned, it is used to respond via :meth:`.client.create_interaction_response`.

    :param client:
        The client to attach the callback to.
    :param callback:
        Used with ``(event)`` if successful or ``(None)`` if timed-out.
    :param component:
        A :code:`'custom_id'` will be set for identifying interactions to it.
    :param timeout:
        Same as :paramref:`.client.wait.timeout`.
    """

    return _interact(client, callback, component, timeout = timeout)




__all__ = (
    'Base', 'Client', 'Connection', 'Interrupted', 'Transmission', 
    'SocketClosed', 'Player', 'Audio', 'Stream', 'StreamPrefixMissing',
    'StreamComplete'
)


class Base(Exception):

    """
    Base module error.
    """

    __slots__ = ()


class Client(Base):

    """
    Base for client errors.
    """

    __slots__ = ()


class Connection(Client):

    """
    Base for connection-related errors.
    """

    __slots__ = ()


class Interrupted(Connection):

    """
    Received a non-salvageable :ddoc:`close code </topics/opcodes-and-status-codes#voice-voice-close-event-codes>`.
    """

    __slots__ = ()


class Transmission(Client):

    """
    Base for creating and sending audio errors.
    """

    __slots__ = ()


class SocketClosed(Transmission):

    """
    The socket is closed upon sending audio.
    """

    __slots__ = ()


class Player(Base):

    """
    Base for :class:`.player.Player` errors.
    """

    __slots__ = ()


class Audio(Player):

    """
    Base for :class:`.audio.Audio` errors.
    """

    __slots__ = ()


class Stream(Audio):

    """
    Base for :class:`.stream.Stream` errors.
    """

    __slots__ = ()


class StreamPrefixMissing(Stream):

    """
    Got an unexpected audio stream OGG prefix.
    """

    __slots__ = ()


class StreamComplete(Stream):

    """
    The audio OGG stream has been exhausted.
    """

    __slots__ = ()



# May 05, 2023

from . import http
from . import gateway
from . import voice
from . import model

from . import enums 
from . import events
from . import client
from . import utils

from . import widget


__all__ = (
    'http', 'gateway', 'voice', 'model', 
    'enums', 'events', 'client', 'utils',
    'widget'
)


import typing
import yarl
import aiohttp
import json
import collections
import asyncio
import types

from . import helpers as _helpers
from . import errors  as _errors


__all__ = ('Client',)


class Client:

    """
    Core means of communication with the HTTP Discord API.

    :param session:
        Used to make requests.
    :param loads:
        Used for converting json text to objects.
    :param dumps:
        Used for converting json objects to text.
    :param ignore_rate:
        Whether to NOT handle rate limiting.
    """

    _version = '10'

    _base_url = yarl.URL(f'https://discord.com/api/v{_version}')

    _user_agent = _helpers.get_user_agent()
    
    _Errors = {
        400: _errors.BadRequest,
        401: _errors.Unauthorized,
        403: _errors.Forbidden,
        404: _errors.NotFound,
        429: _errors.RateLimited
    }

    _rate_marks = {}
    _rate_infos = collections.defaultdict(dict)

    _rate_event = asyncio.Event()
    _rate_event.set()

    __slots__ = ('_session', '_token', '_loads', '_dumps', '_ought_handle_rate')

    def __init__(self, 
                 session    : aiohttp.ClientSession,
                 loads      : typing.Callable[[str], typing.Any] = json.loads,
                 dumps      : typing.Callable[[typing.Any], str] = json.dumps,
                 ignore_rate: bool                               = False):

        self._session = session

        self._token = None

        self._loads = loads
        self._dumps = dumps

        self._ought_handle_rate = not ignore_rate

    def _authenticate(self, token):

        self._token = token

    def authenticate(self,
                     token: None | str) -> None:
        
        """
        Set the authentication token.
        """
        
        self._authenticate(token)

    def _get_headers(self):

        headers = {
            'User-Agent': self._user_agent
        }

        if not (token := self._token) is None:
            headers['Authorization'] = token

        return headers

    async def _execute(self, verb, path, query, data, form, files, headers):

        url = self._base_url.with_path(self._base_url.path + path)

        if not query is None:
            _helpers.clean_query(query)

        sub_headers = headers
            
        headers = self._get_headers()

        if not sub_headers is None:
            headers |= sub_headers

        if not files is None:
            if not form is None:
                raise ValueError('"data" and "files" cannot be used togeter')
            form = aiohttp.FormData(files)
            if not data is None:
                form_json = self._dumps(data)
                form.add_field('payload_json', form_json, content_type = 'application/json')
                data = None

        _helpers.clean_headers(headers)

        response = await self._session.request(
            verb, 
            url, 
            params = query, 
            json = data, 
            data = form,
            headers = headers
        )

        text = await response.text()

        try:
            data = self._loads(text)
        except json.JSONDecodeError:
            data = None

        if response.status < 400:
            return (response, data)

        try:
            Error = self._Errors[response.status]
        except KeyError:
            Error = _errors.Request if response.status < 500 else _errors.Internal

        raise Error(response, data)

    async def _request_bare(self, verb, path, *args, **kwargs):

        response, data = await self._execute(verb, path, *args, **kwargs)

        return data
    
    async def _handle_rate_enter(self, identity):

        await self._rate_event.wait()

        try:
            mark = self._rate_marks[identity]
            info = self._rate_infos[self._token][mark]
        except KeyError:
            return
        
        await info.lock.acquire()

    async def _handle_rate_leave_global(self, headers):

        event = self._rate_event

        if not event.is_set():
            return
        
        event.clear()
        
        loop = asyncio.get_event_loop()

        delay = float(headers['Retry-After'])

        def reset():
            event.set()

        loop.call_later(delay, reset)

    async def _handle_rate_leave_local(self, identity, headers):

        mark = headers['X-RateLimit-Bucket']

        self._rate_marks[identity] = mark

        infos = self._rate_infos[self._token]

        limit = int(headers['X-RateLimit-Limit'])

        try:
            info = infos[mark]
        except KeyError:
            lock = asyncio.Semaphore(value = limit)
            info = infos[mark] = types.SimpleNamespace(lock = lock, timer = None)

        flush = int(headers['X-RateLimit-Remaining'])

        while info.lock._value > flush:
            await info.lock.acquire()

        if flush or not info.timer is None:
            return
        
        loop = asyncio.get_event_loop()

        delay = float(headers['X-RateLimit-Reset-After'])

        def reset():
            for _ in range(limit):
                info.lock.release()
            info.timer = None

        info.timer = loop.call_later(delay, reset)

    def _handle_rate_leave(self, identity, headers):

        score = headers.get('X-RateLimit-Scope', 'user')

        if score == 'global':
            return self._handle_rate_leave_global(headers)

        return self._handle_rate_leave_local(identity, headers)
    
    async def _request_safe(self, identity, verb, path, *args, **kwargs):

        while not self._session.closed:
            await self._handle_rate_enter(identity)
            try:
                response, data = await self._execute(verb, path, *args, **kwargs)
            except _errors.RateLimited as error:
                response = error.response
                delay = error.data['retry_after']
            except:
                delay = None; raise
            else:
                delay = 0   ; break
            finally:
                if not delay is None:
                    await self._handle_rate_leave(identity, response.headers)
                    await asyncio.sleep(delay)
        else:
            raise _errors.Interrupted('session closed while executing a request')

        return data
    
    def _request(self, identity, *args, **kwargs):

        if self._ought_handle_rate and not identity is None:
            return self._request_safe(identity, *args, **kwargs)
        
        return self._request_bare(*args, **kwargs)
        
    async def request(self, 
                      verb    :        str,
                      path    :        str,
                      query   : None | dict[str, str | int | bool] = None, 
                      json    : None | typing.Any                  = None, 
                      data    : None | typing.Any                  = None,
                      files   : None | dict[str, bytes]            = None,
                      headers : None | dict[str, str]              = None,
                      identity: None | typing.Any                  = None) -> typing.Any:
        
        """
        Make a request to the API and get the responding data.

        :param verb:
            The HTTP method.
        :param path:    
            The path portion for the url.
        :param query:
            The HTTP query.
        :param json:
            Used to compose the HTTP body.
        :param data:
            Used to compose the HTTP body.
        :param headers:
            The HTTP headers.
        :param identity:
            Used for handling rate limits.
        """

        data = await self._request(identity, verb, path, query, json, data, files, headers)

        return data




import typing
import asyncio
import time
import itertools

from . import errors as _errors
from . import audio  as _audio


__all__ = ('Player',)



class Player:

    """
    Manages timely sending voice packets.

    :param send:
        Used with ``(samples, data)`` for sending.
    """

    __slots__ = ('_send', '_audio')

    def __init__(self, 
                 send: typing.Callable[[int, bytes], None]):

        self._send = send

        self._audio = None

    @property
    def audio(self) -> _audio.Audio:

        """
        The current audio source.
        """

        return self._audio

    async def _stream(self):
        
        stamp_anchor = time.perf_counter()

        for cycle in itertools.count(1):
            try:
                data = await self._audio.get()
            except _errors.Audio:
                break
            try:
                self._send(self._audio.frame_samples_count, data)
            except _errors.Transmission:
                break
            stamp_actual = time.perf_counter()
            stamp_expect = stamp_anchor + cycle * self._audio.frame_duration
            delay = max(0, stamp_expect - stamp_actual)
            await asyncio.sleep(delay)

    async def _start(self, audio):

        await self._stop()

        await audio.start()

        self._audio = audio

        try:
            await self._stream()
        finally:
            await self._stop()

    def start(self, 
              audio: _audio.Audio) -> typing.Awaitable[None]:
        
        """
        Start sending voice packets.
        
        :param audio:
            Used for reading.

        .. note::
            Will :meth:`~_audio.Audio.close` the source when all packets have been sent.
        """

        return self._start(audio)
    
    async def _stop(self):

        if self._audio is None:
            return

        await self._audio.stop()

    def stop(self) -> typing.Awaitable[None]:

        """
        Close the current audio source.
        """

        return self._stop()
    


import json


__all__ = (
     'Base', 'Client', 'Interrupted', 'Request', 'Internal', 'BadRequest',
     'Unauthorized', 'Forbidden', 'NotFound', 'RateLimited'
)


class Base(Exception):
    
    """
    Base module error.
    """

    __slots__ = ()


class Client(Base):
    
    """
    Base for client errors.
    """

    __slots__ = ()


class Interrupted(Client):

     """
     A request was interrupted.
     """

     __slots__ = ()


class Request(Client):

     """
     Received a response with an unsupported status.
     """

     __slots__ = ('_response', '_data')

     def __init__(self, response, data):

          self._response = response
          self._data = data

          message = json.dumps(data, indent = 2)

          super().__init__(message)

     @property
     def response(self):

          return self._response
     
     @property
     def data(self):

          return self._data


class Internal(Request):

     """
     There was an server error.
     """

     __slots__ = ()


class BadRequest(Request):

     """
     The request was badly formatted.
     """
     
     __slots__ = ()


class Unauthorized(Request):

     """
     The request token is missing or invalid.
     """
     
     __slots__ = ()

    
class Forbidden(Request):

     """
     The user is missing the necessary permissions for the request.
     """
     
     __slots__ = ()


class NotFound(Request):

     """
     The target resource could not be found.
     """
     
     __slots__ = ()


class RateLimited(Request):

     """
     There are too many requests being sent.
     """
     
     __slots__ = ()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import sys
import os
import importlib.metadata

sys.path.insert(0, os.path.abspath('..'))

project = 'aiocord'
author = 'Exahilosys'
copyright = f'2023, {author}'
release = importlib.metadata.version(project)
version = '.'.join(release.split('.')[:2])

metadata = importlib.metadata.metadata('aiocord')

rst_prolog = """
.. |dsrc| replace:: Source:
"""

extensions = [
    'sphinx_rtd_theme',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.extlinks',
    'sphinx_autodoc_typehints',
    'sphinx_paramlinks',
    'sphinx.ext.autosectionlabel'
]

autodoc_typehints = 'signature'

autodoc_member_order = 'bysource'

autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'show-inheritance': True
}

intersphinx_mapping = {
    'python3': ('https://docs.python.org/3', None),
    'aiohttp': ('https://docs.aiohttp.org/en/stable', None),
}

extlinks = {
    'ddoc': ('https://discord.com/developers/docs%s', None)
}

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

paramlinks_hyperlink_param = 'name'

html_theme = 'basic'

html_static_path = ['_static']

html_css_files = [
    'main.css'
]

pygments_style = 'stata-dark'

html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html']
}

html_show_copyright = False

html_show_sphinx = False

html_theme_options = {
    'globaltoc_includehidden': True,
    'globaltoc_maxdepth': -1
}

html_context = {
    'github_url': metadata['Home-Page']
}



__all__ = ()


def clean_query(store):

    for key in tuple(store):
        value = store[key]
        if value is None:
            del store[key]; continue
        if isinstance(value, bool):
            value = int(value)
        if isinstance(value, int):
            value = str(value)
        store[key] = value


def clean_headers(store):

    for key in tuple(store):
        value = store[key]
        if value is None:
            del store[key]; continue
        store[key] = str(value)


import typing
import aiohttp
import json
import asyncio
import socket
import struct
import nacl.secret

from . import helpers as _helpers
from . import enums   as _enums
from . import errors  as _errors
from . import vital   as _vital
from . import player  as _player


__all__ = ('Client',)


class Client:

    """
    Core means of communication with the Discord Voice API.

    :param callback:
        Used for dispatching events.
    :param session:
        Used for creating websocket connections.
    :param user_id:
        The connecting client's user id.
    :param guild_id:
        The connecting guild's id.
    :param session_id:
        The :ddoc:`voice state </topics/gateway-events#voice-state-update>` session id.
    :param token:
        The :ddoc:`voice server </topics/gateway-events#voice-state-update>` token.
    :param uri:
        The :ddoc:`voice server </topics/gateway-events#voice-state-update>` endpoint.
    :param loads:
        Used for converting json text to objects.
    :param dumps:
        Used for converting json objects to text.
    """

    _version = 4

    __slots__ = (
        '_callback', '_session', '_user_id', '_guild_id', '_session_id', 
        '_token', '_uri', '_loads', '_dumps', '_websocket', '_vital',
        '_event_identify', '_event_complete', '_ssrc', '_address', '_socket',
        '_encrypt', '_sequence', '_timestamp', '_secret', '_codec', '_player'
    )

    def __init__(self, 
                 callback  : typing.Callable[[str, typing.Any], typing.Awaitable[None]],
                 session   : aiohttp.ClientSession, 
                 user_id   : str, 
                 guild_id  : str, 
                 session_id: str, 
                 token     : str, 
                 uri       : str,
                 loads     : typing.Callable[[str], typing.Any] = json.loads,
                 dumps     : typing.Callable[[typing.Any], str] = json.dumps):

        self._callback = callback

        self._session = session

        self._user_id = user_id
        self._guild_id = guild_id
        self._session_id = session_id
        self._token = token

        self._uri = uri

        self._loads = loads
        self._dumps = dumps

        self._websocket = None

        self._vital = _vital.Vital(self._vital_beat, self._vital_save)

        self._event_identify = asyncio.Event()
        self._event_complete = asyncio.Event()

        self._ssrc = None
        self._address = None

        self._socket = None

        self._encrypt = None

        self._sequence = 0
        self._timestamp = 0

        self._secret = None
        self._codec = None

        self._player = _player.Player(self._send_audio)

    @property
    def guild_id(self) -> str:

        """
        The connected guild's id.
        """

        return self._guild_id

    @property
    def event_identify(self) -> asyncio.Event:

        """
        Set after the identify payload has been sent.
        """

        return self._event_identify
    
    @property
    def event_complete(self) -> asyncio.Event:

        """
        Set after the session description has been handled or when resumed.
        """

        return self._event_complete
    
    @property
    def codec(self) -> str:

        """
        The name of the expected codec.
        """

        return self._codec
    
    @property
    def player(self) -> _player.Player:

        """
        The player.
        """

        return self._player

    async def _recv_packet(self, size):

        loop = asyncio.get_event_loop()

        data = await loop.sock_recv(self._socket, size)

        return data
    
    def _send_packet(self, data):

        if self._socket._closed:
            raise _errors.SocketClosed()

        size = self._socket.send(data)

        return size
    
    def _make_packet(self, data):

        header = self._make_packet_header()

        data = self._encrypt(bytes(header), data)

        return header + data
    
    def _make_packet_header(self):

        header = bytearray(12)

        header[0] = 0x80
        header[1] = 0x78

        struct.pack_into('>H', header, 2, self._sequence)
        struct.pack_into('>I', header, 4, self._timestamp)
        struct.pack_into('>I', header, 8, self._ssrc)

        return header
    
    def _send_audio(self, samples, data):

        packet = self._make_packet(data)
        
        self._send_packet(packet)

        self._sequence = _helpers.max_add(self._sequence, 1, 65535)
        self._timestamp = _helpers.max_add(self._timestamp, samples, 4294967295)

    def send_audio(self, samples: int, data: bytes) -> None:

        """
        Send an audio packet.
        
        :param samples:
            The sample count per frame.
        :param data:
            The encoded packet data.
        """

        return self._send_audio(samples, data)

    async def _send(self, code, data):

        info = {'op': code, 'd': data}

        await self._websocket.send_json(info, dumps = self._dumps)

    async def _send_resume(self):

        code = _enums.OpCode.resume

        data = {
            'server_id': self._guild_id,
            'session_id': self._session_id,
            'token': self._token
        }

        await self._send(code, data)

    async def _send_speaking(self, flags):

        code = _enums.OpCode.speaking

        data = {
            'speaking': flags,
            'delay': 0,
            'ssrc': self._ssrc
        }

        await self._send(code, data)

    async def _send_select_protocol(self, ip, port, mode):

        code = _enums.OpCode.select_protocol

        data = {
            'protocol': 'udp',
            'data': {
                'address': ip,
                'port': port,
                'mode': mode
            }
        }

        await self._send(code, data)
        
    async def _send_heartbeat(self):

        loop = asyncio.get_event_loop()

        code = _enums.OpCode.heartbeat

        data = loop.time()

        await self._send(code, data)

    async def _send_identify(self):

        code = _enums.OpCode.identify

        data = {
            'server_id': self._guild_id,
            'user_id': self._user_id,
            'session_id': self._session_id,
            'token': self._token
        }

        await self._send(code, data)

    async def _resume(self):

        self._event_complete.clear()

        await self._disconnect()

        await self._send_resume()

        await self._event_complete.wait()

    async def _vital_save(self):

        await self._resume()

    async def _vital_beat(self):

        await self._send_heartbeat()

    async def _discover_address(self):
        
        size = 74

        packet = bytearray(size)

        struct.pack_into('>H', packet, 0, 1)
        struct.pack_into('>H', packet, 2, 70)
        struct.pack_into('>I', packet, 4, self._ssrc)

        self._send_packet(packet)

        data = await self._recv_packet(size)

        ip_enter = 8
        ip_leave = data.index(0, ip_enter)
        ip = data[ip_enter:ip_leave]
        ip = ip.decode()

        port = struct.unpack_from('>H', data, size - 2)[0]
        
        return ip, port

    async def _connect_udp(self):

        loop = asyncio.get_event_loop()

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        sock.setblocking(False)

        await loop.sock_connect(sock, self._address)

        self._socket = sock

    def _encrypt_xsalsa20_poly1305(self, header, data):

        nonce = bytearray(24)

        nonce[:12] = header

        message = self._secret.encrypt(data, bytes(nonce))

        return message.ciphertext
    
    async def _handle_disconnect(self, info, data):

        name = 'LEAVE'

        await self._callback(name, data)
    
    async def _handle_connect(self, info, data):

        name = 'ENTER'

        await self._callback(name, data)
    
    async def _handle_speaking(self, info, data):

        name = 'SPEAK'

        await self._callback(name, data)
    
    async def _handle_resumed(self, info, data):

        self._event_complete.set()

    async def _handle_session_description(self, info, data):

        mode = data['mode']

        self._encrypt = getattr(self, f'_encrypt_{mode}')

        self._secret = nacl.secret.SecretBox(bytes(data['secret_key']))

        self._codec = data['audio_codec']

        await self._send_speaking(1)

        self._event_complete.set()

    async def _handle_ready(self, info, data):

        self._ssrc = data['ssrc']

        self._address = (data['ip'], data['port'])

        for mode in data['modes']:
            if not hasattr(self, f'_encrypt_{mode}'):
                continue
            break
        else:
            raise RuntimeError('missing supported mode')

        await self._connect_udp()

        ip, port = await self._discover_address()

        await self._send_select_protocol(ip, port, mode)

    async def _handle_heartbeat_ack(self, info, data):

        self._vital.ack()

    async def _handle_hello(self, info, data):
        
        self._vital.start(data['heartbeat_interval'] / 1000)

    async def _listen_TEXT(self, data):

        info = self._loads(data)

        try:
            code = _enums.OpCode(info.pop('op'))
        except ValueError:
            return

        try:
            handle = getattr(self, f'_handle_{code.name}')
        except AttributeError:
            return # warn log
        
        data = info.pop('d', None)
        
        await handle(info, data)

    _close_codes_finish = {None, 1000}

    _close_codes_resume = {4015}

    async def _egress(self):

        close_code = self._websocket.close_code

        if close_code in self._close_codes_finish:
            return
        
        for routine in (self._resume,):
            close_codes_name =  f'_clode_codes_{routine.__name__}'
            close_codes = getattr(self, close_codes_name)
            if not close_code in close_codes:
                continue
            break
        else:
            raise _errors.Interrupted()
        
        await routine()

    async def _listen(self):

        async for message in self._websocket:
            try:
                handle = getattr(self, f'_listen_{message.type.name}')
            except AttributeError:
                continue
            await handle(message.data)

    async def _disconnect(self):

        await self._websocket.close(code = 1000)
    
    async def _connect(self):

        query = {
            'v': self._version
        }

        self._websocket = await self._session.ws_connect(self._uri, params = query)

    async def _start(self):

        await self._connect()

        await self._send_identify()

        self._event_identify.set()

        await self._listen()

    def start(self) -> typing.Awaitable[None]:

        """
        Start the :ddoc:`connection lifecycle </topics/voice-connections#connecting-to-voice>`.
        """

        return self._start()
    
    async def _stop(self):

        await self._player.stop()

        self._socket.close()

        self._vital.stop()

        await self._websocket.close()

    def stop(self) -> typing.Awaitable[None]:

        """
        Cancel tasks and close connections.
        """

        return self._stop()
    



import enum


__all__ = ('Intents', 'GatewayEvent', 'VoiceEvent')


class Intents(enum.IntFlag):

    """
    Available :ddoc:`Gateway Intents </topics/gateway#gateway-intents>`.
    """
  
    guilds                           = 1 << 0
    guild_members                    = 1 << 1
    guild_moderation                 = 1 << 2
    guild_emojis_and_stickers        = 1 << 3
    guild_integrations               = 1 << 4
    guild_webhooks                   = 1 << 5
    guild_invites                    = 1 << 6
    guild_voice_states               = 1 << 7
    guild_presences                  = 1 << 8
    guild_messages                   = 1 << 9
    guild_message_reactions          = 1 << 10
    guild_message_typing             = 1 << 11
    direct_messages                  = 1 << 12
    direct_message_reactions         = 1 << 13
    direct_message_typing            = 1 << 14
    message_content                  = 1 << 15
    guild_scheduled_events           = 1 << 16
    auto_moderation_configuration    = 1 << 20
    auto_moderation_action_execution = 1 << 21

    @classmethod
    def default(cls):

        """
        Get all the non-privilaged intents.
        """

        return (
              cls.guilds
            | cls.guild_moderation
            | cls.guild_emojis_and_stickers
            | cls.guild_integrations
            | cls.guild_webhooks
            | cls.guild_invites
            | cls.guild_voice_states
            | cls.guild_messages
            | cls.guild_message_reactions
            | cls.guild_message_typing
            | cls.direct_messages
            | cls.direct_message_reactions
            | cls.direct_message_typing
            | cls.guild_scheduled_events
            | cls.auto_moderation_configuration
            | cls.auto_moderation_action_execution
        )


class GatewayEvent(enum.StrEnum):

    """
    Available `Events </topics/gateway-events#receive-events>` received through :attr:`.gateway.enums.OpCode.dispatch`.
    """

    ready                                  = 'READY'
    """
    |dsrc| 
    :ddoc:`Ready </topics/gateway-events#ready>`
    """
    update_application_command_permissions = 'APPLICATION_COMMAND_PERMISSIONS_UPDATE'
    """
    |dsrc| 
    :ddoc:`Application Command Permissions Update <topics/gateway-events#application-command-permissions-update>`
    """
    create_auto_moderation_rule            = 'AUTO_MODERATION_RULE_CREATE'
    """
    |dsrc| 
    :ddoc:`Auto Moderation Rule Create </topics/gateway-events#auto-moderation-rule-create>`
    """
    update_auto_moderation_rule            = 'AUTO_MODERATION_RULE_UPDATE'
    """
    |dsrc| 
    :ddoc:`Auto Moderation Rule Update </topics/gateway-events#auto-moderation-rule-update>`
    """
    delete_auto_moderation_rule            = 'AUTO_MODERATION_RULE_DELETE'
    """
    |dsrc| 
    :ddoc:`Auto Moderation Rule Delete </topics/gateway-events#auto-moderation-rule-delete>` 
    """
    execute_auto_moderation_action         = 'AUTO_MODERATION_ACTION_EXECUTION'
    """
    |dsrc| 
    :ddoc:`Auto Moderation Action Execution </topics/gateway-events#auto-moderation-action-execution>` 
    """
    create_channel                         = 'CHANNEL_CREATE'
    """
    |dsrc| 
    :ddoc:`Channel Create </topics/gateway-events#channel-create>` 
    """
    update_channel                         = 'CHANNEL_UPDATE'
    """
    |dsrc| 
    :ddoc:`Channel Update </topics/gateway-events#channel-update>` 
    """
    delete_channel                         = 'CHANNEL_DELETE'
    """
    |dsrc| 
    :ddoc:`Channel Delete </topics/gateway-events#channel-delete>` 
    """
    update_channel_pins                    = 'CHANNEL_PINS_UPDATE'
    """
    |dsrc| 
    :ddoc:`Channel Pins Update </topics/gateway-events#channel-pins-update>` 
    """
    create_thread                          = 'THREAD_CREATE'
    """
    |dsrc| 
    :ddoc:`Thread Create </topics/gateway-events#thread-create>` 
    """
    update_thread                          = 'THREAD_UPDATE'
    """
    |dsrc| 
    :ddoc:`Thread Update </topics/gateway-events#thread-update>` 
    """
    delete_thread                          = 'THREAD_DELETE'
    """
    |dsrc| 
    :ddoc:`Thread Delete </topics/gateway-events#thread-delete>` 
    """
    sync_threads                           = 'THREAD_LIST_SYNC'
    """
    |dsrc| 
    :ddoc:`Thread List Sync </topics/gateway-events#thread-list-sync>` 
    """
    update_thread_member                   = 'THREAD_MEMBER_UPDATE'
    """
    |dsrc| 
    :ddoc:`Thread Member Update </topics/gateway-events#thread-member-update>` 
    """
    update_thread_members                  = 'THREAD_MEMBERS_UPDATE'
    """
    |dsrc| 
    :ddoc:`Thread Members Update </topics/gateway-events#thread-members-update>` 
    """
    create_guild                           = 'GUILD_CREATE'
    """
    |dsrc| 
    :ddoc:`Guild Create </topics/gateway-events#guild-create>` 
    """
    update_guild                           = 'GUILD_UPDATE'
    """
    |dsrc| 
    :ddoc:`Guild Update </topics/gateway-events#guild-update>` 
    """
    delete_guild                           = 'GUILD_DELETE'
    """
    |dsrc| 
    :ddoc:`Guild Delete </topics/gateway-events#guild-delete>` 
    """
    create_guild_audit_log_entry           = 'GUILD_AUDIT_LOG_ENTRY_CREATE'
    """
    |dsrc| 
    :ddoc:`Guild Audit Log Entry Create </topics/gateway-events#guild-audit-log-entry-create>` 
    """
    create_guild_ban                       = 'GUILD_BAN_ADD'
    """
    |dsrc| 
    :ddoc:`Guild Ban Add </topics/gateway-events#guild-ban-add>` 
    """
    delete_guild_ban                       = 'GUILD_BAN_REMOVE'
    """
    |dsrc| 
    :ddoc:`Guild Ban Remove </topics/gateway-events#guild-ban-remove>` 
    """
    update_guild_emojis                    = 'GUILD_EMOJIS_UPDATE'
    """
    |dsrc| 
    :ddoc:`Guild Emojis Update </topics/gateway-events#guild-emojis-update>` 
    """
    update_guild_stickers                  = 'GUILD_STICKERS_UPDATE'
    """
    |dsrc| 
    :ddoc:`Guild Stickers Update </topics/gateway-events#guild-stickers-update>` 
    """
    update_guild_integrations              = 'GUILD_INTEGRATIONS_UPDATE'
    """
    |dsrc| 
    :ddoc:`Guild Integrations Update </topics/gateway-events#guild-integrations-update>` 
    """
    create_guild_member                    = 'GUILD_MEMBER_ADD'
    """
    |dsrc| 
    :ddoc:`Guild Member Add </topics/gateway-events#guild-member-add>` 
    """
    delete_guild_member                    = 'GUILD_MEMBER_REMOVE'
    """
    |dsrc| 
    :ddoc:`Guild Member Remove </topics/gateway-events#guild-member-remove>` 
    """
    update_guild_member                    = 'GUILD_MEMBER_UPDATE'
    """
    |dsrc| 
    :ddoc:`Guild Member Update </topics/gateway-events#guild-member-update>` 
    """
    receive_guild_members                  = 'GUILD_MEMBERS_CHUNK'
    """
    |dsrc| 
    :ddoc:`Guild Members Chunk </topics/gateway-events#guild-members-chunk>` 
    """
    create_guild_role                      = 'GUILD_ROLE_CREATE'
    """
    |dsrc| 
    :ddoc:`Guild Role Create </topics/gateway-events#guild-role-create>` 
    """
    update_guild_role                      = 'GUILD_ROLE_UPDATE'
    """
    |dsrc| 
    :ddoc:`Guild Role Update </topics/gateway-events#guild-role-update>` 
    """
    delete_guild_role                      = 'GUILD_ROLE_DELETE'
    """
    |dsrc| 
    :ddoc:`Guild Role Delete </topics/gateway-events#guild-role-delete>` 
    """
    create_guild_scheduled_event           = 'GUILD_SCHEDULED_EVENT_CREATE'
    """
    |dsrc| 
    :ddoc:`Guild Scheduled Event Create </topics/gateway-events#guild-scheduled-event-create>` 
    """
    update_guild_scheduled_event           = 'GUILD_SCHEDULED_EVENT_UPDATE'
    """
    |dsrc| 
    :ddoc:`Guild Scheduled Event Update </topics/gateway-events#guild-scheduled-event-update>` 
    """
    delete_guild_scheduled_event           = 'GUILD_SCHEDULED_EVENT_DELETE'
    """
    |dsrc| 
    :ddoc:`Guild Scheduled Event Delete </topics/gateway-events#guild-scheduled-event-delete>` 
    """
    create_guild_scheduled_event_user      = 'GUILD_SCHEDULED_EVENT_USER_ADD'
    """
    |dsrc| 
    :ddoc:`Guild Scheduled Event User Add </topics/gateway-events#guild-scheduled-event-user-add>` 
    """
    delete_guild_scheduled_event_user      = 'GUILD_SCHEDULED_EVENT_USER_REMOVE'
    """
    |dsrc| 
    :ddoc:`Guild Scheduled Event User Remove </topics/gateway-events#guild-scheduled-event-user-remove>` 
    """
    create_integration                     = 'INTEGRATION_CREATE'
    """
    |dsrc| 
    :ddoc:`Integration Create </topics/gateway-events#integration-create>` 
    """
    update_integration                     = 'INTEGRATION_UPDATE'
    """
    |dsrc| 
    :ddoc:`Integration Update </topics/gateway-events#integration-update>` 
    """
    delete_integration                     = 'INTEGRATION_DELETE'
    """
    |dsrc| 
    :ddoc:`Integration Delete </topics/gateway-events#integration-delete>` 
    """
    create_interaction                     = 'INTERACTION_CREATE'
    """
    |dsrc| 
    :ddoc:`Interaction Create </topics/gateway-events#interaction-create>` 
    """
    create_invite                          = 'INVITE_CREATE'
    """
    |dsrc| 
    :ddoc:`Invite Create </topics/gateway-events#invite-create>` 
    """
    delete_invite                          = 'INVITE_DELETE'
    """
    |dsrc| 
    :ddoc:`Invite Delete </topics/gateway-events#invite-delete>` 
    """
    create_message                         = 'MESSAGE_CREATE'
    """
    |dsrc| 
    :ddoc:`Message Create </topics/gateway-events#message-create>` 
    """
    update_message                         = 'MESSAGE_UPDATE'
    """
    |dsrc| 
    :ddoc:`Message Update </topics/gateway-events#message-update>` 
    """
    delete_message                         = 'MESSAGE_DELETE'
    """
    |dsrc| 
    :ddoc:`Message Delete </topics/gateway-events#message-delete>` 
    """
    delete_messages                        = 'MESSAGE_DELETE_BULK'
    """
    |dsrc| 
    :ddoc:`Message Delete Bulk </topics/gateway-events#message-delete-bulk>` 
    """
    create_message_reaction                = 'MESSAGE_REACTION_ADD'
    """
    |dsrc| 
    :ddoc:`Message Reaction Add </topics/gateway-events#message-reaction-add>` 
    """
    delete_message_reaction                = 'MESSAGE_REACTION_REMOVE'
    """
    |dsrc| 
    :ddoc:`Message Reaction Remove </topics/gateway-events#message-reaction-remove>` 
    """
    delete_all_message_reactions           = 'MESSAGE_REACTION_REMOVE_ALL'
    """
    |dsrc| 
    :ddoc:`Message Reaction Remove All </topics/gateway-events#message-reaction-remove-all>` 
    """
    delete_all_message_emoji_reactions     = 'MESSAGE_REACTION_REMOVE_EMOJI'
    """
    |dsrc| 
    :ddoc:`Message Reaction Remove Emoji </topics/gateway-events#message-reaction-remove-emoji>` 
    """
    update_presence                        = 'PRESENCE_UPDATE'
    """
    |dsrc| 
    :ddoc:`Presence Update </topics/gateway-events#presence-update>` 
    """
    create_stage_instance                  = 'STAGE_INSTANCE_CREATE'
    """
    |dsrc| 
    :ddoc:`Stage Instance Create </topics/gateway-events#stage-instance-create>` 
    """
    update_stage_instance                  = 'STAGE_INSTANCE_UPDATE'
    """
    |dsrc| 
    :ddoc:`Stage Instance Update </topics/gateway-events#stage-instance-update>` 
    """
    delete_stage_instance                  = 'STAGE_INSTANCE_DELETE'
    """
    |dsrc| 
    :ddoc:`Stage Instance Delete </topics/gateway-events#stage-instance-delete>` 
    """
    create_typing_indicator                = 'TYPING_START'
    """
    |dsrc| 
    :ddoc:`Typing Start </topics/gateway-events#typing-start>` 
    """
    update_self_user                       = 'USER_UPDATE'
    """
    |dsrc| 
    :ddoc:`User Update </topics/gateway-events#user-update>` 
    """
    update_voice_state                     = 'VOICE_STATE_UPDATE'
    """
    |dsrc| 
    :ddoc:`Voice State Update </topics/gateway-events#voice-state-update>` 
    """
    update_voice_server                    = 'VOICE_SERVER_UPDATE'
    """
    |dsrc| 
    :ddoc:`Voice Server Update </topics/gateway-events#voice-server-update>` 
    """
    update_webhooks                        = 'WEBHOOKS_UPDATE'
    """
    |dsrc| 
    :ddoc:`Webhooks Update </topics/gateway-events#webhooks-update>` 
    """


class VoiceEvent(enum.StrEnum):

    speak = 'SPEAK'
    """
    A user has spoken.
    """
    enter = 'ENTER'
    """
    A user has joined the channel.
    """
    leave = 'LEAVE'
    """
    A user has left the channel.
    """


import typing
import os
import asyncio
import functools
import types
import weakref
import inspect

from . import helpers as _helpers
from . import model   as _model
from . import client  as _client
from . import events  as _events


__all__ = ('Info', 'callback', 'interact', 'load', 'drop')


_V = typing.TypeVar('_V')
_P = typing.ParamSpec('_P')


class Info(typing.NamedTuple):

    """
    Holds global information.
    """

    client : _client.Client
    """
    The client for which this widget has been loaded.
    """
    widgets: dict[str, types.ModuleType]
    """
    All currently loaded widgets.
    """


def _load_module(name, path, intermediary = '_m_'):

    module_name = f'{__name__}.{intermediary}.{name}'

    module = _helpers.load_module_via_path(module_name, path)

    return module


_latent_callback_assets = []


def _callback(Event):
    def decorator(function):
        _latent_callback_assets.append((Event, function))
        return function
    return decorator


def callback(Event: _V) -> typing.Callable[[typing.Callable[_P, _V]], typing.Callable[[Info, typing.Unpack[tuple[_V, ...]]], None]]:

    """
    Create a callback for an core_event.

    :param Event:
        Any event dispatchable by :class:`.client.Client`.
    """

    return _callback(Event)


_latent_interact_assets = []


def _interact(path):
    def decorator(function):
        _latent_interact_assets.append((path, function))
        return function
    return decorator


def interact(*path: typing.Unpack[tuple[str, ...]]) -> typing.Callable[[typing.Callable[_P, _V]], typing.Callable[[Info, _events.CreateIntegration], typing.Awaitable[_model.protocols.InteractionResponse | None]]]:

    """
    Create a callback for an interaction.

    If a :code:`.model.protocols.InteractionResponse` is returned, it is used to respond via :meth:`.client.create_interaction_response`.

    :param *path:
        The ``command-name`` to ``sub[command/group]-name`` leading to the desired command.
    """

    return _interact(path)


_modules_group = weakref.WeakKeyDictionary()


_assets = weakref.WeakKeyDictionary()


def _find_caller_widget_asset():

    stack = inspect.stack()

    for index, frame in enumerate(stack):
        if index < 2:
            continue
        module = inspect.getmodule(frame)
        try:
            asset = _assets[module]
        except KeyError:
            continue
        return asset
    
    return None


async def _load(client, name, path):

    if path is None:
        asset = _find_caller_widget_asset()
        path = '.' if asset is None else asset.path

    loop = asyncio.get_event_loop()

    try:
        modules = _modules_group[client]
    except KeyError:
        modules = _modules_group[client] = weakref.WeakValueDictionary()

    try:
        await _drop(client, name, pop = False)
    except KeyError:
        pass

    if not (application := client.cache.application):
        application = await client.get_self_application_information()

    commands = await client.get_global_application_commands(application.id)

    module_path = os.path.join(path, name)
    
    load_module = functools.partial(_load_module, name, module_path, intermediary = id(client))

    module = modules[name] = await loop.run_in_executor(None, load_module)

    # module.__name__, name = name, module.__name__

    info = Info(client, types.MappingProxyType(modules))

    callbacks = []

    Events = set()

    def load_callback(Event, function):
        Events.add(Event)
        async def callback(event, *args, **kwargs):
            if not isinstance(event, Event):
                return
            await function(info, event, *args, **kwargs)
        return callback

    for asset in _latent_callback_assets:
        callback = load_callback(*asset)
        callbacks.append(callback)

    _latent_callback_assets.clear()

    def get_command(path, commands = commands):
        name, *path = path
        for command in commands:
            if not command.name == name:
                continue
            break
        if not path:
            return command
        return get_command(path, commands = command.options)

    def load_interact_callback(path, function):
        command = get_command(path)
        async def callback(info, core_event, *args, **kwargs):
            if not core_event.interaction.type == _model.enums.InteractionType.application_command:
                return
            if not core_event.interaction.data.id == command.id:
                return
            response = await function(info, core_event, *args, **kwargs)
            if not response is None:
                await client.create_interaction_response(core_event.interaction.id, core_event.interaction.token, **response)
        return callback
    
    for asset in _latent_interact_assets:
        callback = load_interact_callback(*asset)
        callback = load_callback(_events.CreateInteraction, callback)
        callbacks.append(callback)

    _latent_interact_assets.clear()

    try:
        load = getattr(module, '__load__')
    except AttributeError:
        pass
    else:
        await load(info)

    client.callbacks.extend(callbacks)

    _assets[module] = types.SimpleNamespace(
        name = name, 
        path = path, 
        info = info, 
        Events = Events,
        callbacks = callbacks
    )

    return module


def load(client: _client.Client, 
         name  : str, 
         path  : str = None) -> typing.Awaitable[types.ModuleType]:
    
    """
    Load a widget by creating and attaching events to the client.
    
    :param client:
        The client to load the widget for.
    :param name:
        The name of the widget, used for identifying in :attr:`.Info.widgets` and :func:`.drop`\ing.
    :param path:
        The location of the module's parent directory. If not specified, the caller widget's parent directory is used.

    The widget may define a ``__load__(info)`` function, which will be called before callbacks are attached.
    """

    return _load(client, name, path)


def _drop_module(name):

    _helpers.drop_module(name)


async def _drop(client, name, *, pop = True):

    loop = asyncio.get_event_loop()

    modules = _modules_group[client]

    getter = modules.pop if pop else modules.__getitem__

    try:
        module = getter(name)
    except KeyError:
        return

    asset = _assets.pop(module)

    for callback in asset.callbacks:
        client.callbacks.remove(callback)

    try:
        drop = getattr(module, '__drop__')
    except AttributeError:
        pass
    else:
        await drop(asset.info)

    drop_module = functools.partial(_drop_module, asset.name)

    await loop.run_in_executor(None, drop_module)


def drop(client: _client.Client, 
         name: str) -> typing.Awaitable[None]:

    """
    Unload a widget by detaching all the related callbles from the client.

    :param client:
        The client to unload the widget for.
    :param name:
        The name of the widget, as specified in :func:`.load`.

    The widget may define a ``__drop__`` function which will be called after callbacks are detached.
    """

    return _drop(client, name)



import typing
import json
import aiohttp
import collections
import zlib
import random
import os
import asyncio

from . import helpers as _helpers
from . import errors  as _errors
from . import enums   as _enums
from . import vital   as _vital


__all__ = ('Client',)


_Info = collections.namedtuple('Info', 'id count')


class Client:

    """
    Core means of communication with the Discord Gateway API.

    :param callback:
        Used for dispatching events.
    :param session:
        Used for creating websocket connections.
    :param token:
        Used for authentication during 
        :ddoc:`identify </topics/gateway-events#identify>`
        and
        :ddoc:`resume </topics/gateway-events#identify>`.
    :param intents:
        Used for signaling which events to allow through during
        :ddoc:`identify </topics/gateway-events#identify>`.
    :param uri:
        Used for creating werbsocket connections. Must be fetched from
        `Get Gateway:ddoc: Bot </topics/gateway#get-gateway-bot>`.
    :param info_id:
        The id of the shard.
    :param info_count:
        The total amount of shards.
    :param encoding:
        The desired
        :ddoc:`encoding </topics/gateway#encoding-and-compression>`.
        Can be :code:`'json'` or :code:`'etf'`, and by chosing either, 
        :paramref:`.loads` and :paramref:`.dumps` should be adjusted accordingly.
    :param present:
        Used for fetching the initial presence to send during 
        :ddoc:`identify </topics/gateway-events#identify>`.
    :param loads:
        Used for converting json text to objects.
    :param dumps:
        Used for converting json objects to text.
    """

    _version = 10
    _zlib_suffix = b'\x00\x00\xff\xff'

    __slots__ = (
        '_session', '_token', '_intents', '_uri', '_info', '_websocket', 
        '_buffer', '_flator', '_vital', '_sequence', '_callback', '_encoding',
        '_present', '_session_id', '_resume_uri', '_event_identify', '_event_complete',
        '_loads', '_dumps'
    )

    def __init__(self, 
                 callback  : typing.Callable[[str, typing.Any], typing.Awaitable[None]],
                 session   : aiohttp.ClientSession, 
                 token     : str,
                 intents   : int,
                 uri       : str, 
                 info_id   : int, 
                 info_count: int, 
                 encoding  : str                                                = 'json',
                 present   : None | typing.Callable[[], typing.Awaitable[dict]] = None,
                 loads     : typing.Callable[[str], typing.Any]                 = json.loads,
                 dumps     : typing.Callable[[typing.Any], str]                 = json.dumps):

        self._callback = callback

        self._session = session

        self._token = token
        self._intents = intents
        self._uri = uri
        self._info = _Info(info_id, info_count)

        self._websocket = None

        self._buffer = bytearray()
        self._flator = zlib.decompressobj()

        self._vital = _vital.Vital(self._vital_beat, self._vital_save)
        self._sequence = None

        self._encoding = encoding

        self._present = present

        self._session_id = None
        self._resume_uri = None
        
        self._event_identify = asyncio.Event()
        self._event_complete = asyncio.Event()

        self._loads = loads
        self._dumps = dumps

    @property
    def info(self):

        return self._info
    
    @property
    def event_identify(self):

        return self._event_identify
    
    @property
    def event_complete(self):

        return self._event_complete
    
    def _inform_session(self, session_id, resume_uri):

        self._session_id = session_id
        self._resume_uri = resume_uri

    def inform_session(self, 
                       session_id: str, 
                       resume_uri: str):

        """
        Update the session information.

        :param session_id:
            The session id. Can be found as :code:`'session_id'` in the 
            :ddoc:`ready </topics/gateway#ready-event>` event.
        :param resume_uri:
            Used for resuming the session. Can be found as :code:`'resume_gateway_url'` in the 
            :ddoc:`ready </topics/gateway#ready-event>` event.
        """

        self._inform_session(session_id, resume_uri)
    
    async def _send(self, code, data):

        info = {'op': code, 'd': data}

        await self._websocket.send_json(info, dumps = self._dumps)

    async def _send_request_guild_members(self, data):
        
        code = _enums.OpCode.request_guild_members

        await self._send(code, data)

    async def request_guild_members(self, data):

        """
        |dsrc| 
        :ddoc:`Request Guild Members </topics/gateway-events#request-guild-members>`
        """

        await self._send_request_guild_members(data)

    async def _send_update_voice_state(self, data):

        code = _enums.OpCode.update_voice_state

        await self._send(code, data)

    def update_voice_state(self, data) -> typing.Awaitable[None]:
        
        """
        |dsrc| 
        :ddoc:`Update Voice State </topics/gateway-events#update-voice-state>`
        """

        return self._send_update_voice_state(data)

    async def _send_update_presence(self, data):

        code = _enums.OpCode.update_presence

        await self._send(code, data)

    def update_presence(self, 
                        status    : str, 
                        activities: list[dict], 
                        since     : int | None, 
                        afk       : bool) -> typing.Awaitable[None]:

        """
        |dsrc|
        :ddoc:`Update Presence </topics/gateway-events#update-presence>`
        """

        return self._send_update_presence(since, activities, status, afk)

    async def _send_resume(self):

        code = _enums.OpCode.resume

        data = {
            'token': self._token,
            'session_id': self._session_id,
            'seq': self._sequence
        }

        await self._send(code, data)

    async def _send_heartbeat(self):

        code = _enums.OpCode.heartbeat

        data = self._sequence

        await self._send(code, data)

    async def _send_identify(self):

        code = _enums.OpCode.identify

        system_name = os.name
        module_name = __name__.split('.', 1)[0]

        data = {
            'token': self._token,
            'properties': {
                'os': system_name,
                'browser': module_name,
                'device': module_name
            },
            'shard': self._info,
            'token': self._token,
            'intents': self._intents
        }

        if not (present := self._present) is None:
            data['presence'] = await present()

        await self._send(code, data)

    async def _resume(self):

        self._event_complete.clear()

        await self._connect(self._resume_uri)

        await self._send_resume()

        await self._event_complete.wait()

    async def _vital_beat(self):

        await self._send_heartbeat()

    async def _vital_save(self):

        await self._disconnect()

        await self._resume()

    async def _handle_invalid_session(self, info, data):

        resumeable = info.get('d', False)

        if resumeable:
            await self._resume(); return
        
        resuming = self._websocket._response.url == self._resume_uri

        if resuming:
            await self._restart(); return

        raise _errors.Invalidated()

    async def _handle_reconnect(self, info, data):

        await self._resume()

    async def _handle_dispatch(self, info, data):

        self._event_complete.set()
        
        self._sequence = info['s']

        name = info['t']

        await self._callback(name, data)
    
    async def _handle_heartbeat(self, info, data):

        await self._vital_beat()

    async def _handle_heartbeat_ack(self, info, data):

        self._vital.ack()
    
    async def _handle_hello(self, info, data):

        interval = data['heartbeat_interval'] / 1000

        def jitter(cycle):
            if cycle:
                return 1
            return random.random()

        self._vital.start(interval, jitter = jitter)

        await self._send_identify()

        self._event_identify.set()

    async def _listen_BINARY(self, data):
        
        self._buffer.extend(data)

        if len(data) < 4 or not data.endswith(self._zlib_suffix):
            return
        
        info = self._flator.decompress(self._buffer)

        self._buffer.clear()

        info = info.decode()
        info = self._loads(info)

        code = _enums.OpCode(info.pop('op'))

        try:
            handle = getattr(self, f'_handle_{code.name}')
        except AttributeError:
            return # warn log
        
        data = info.pop('d', None)
    
        await handle(info, data)

    _close_codes_finish = {None, 1000}

    _close_codes_resume = {4000, 4002, 4003, 4005, 4007, 4008, 4009}

    async def _egress(self):

        close_code = self._websocket.close_code

        if close_code in self._close_codes_finish:
            return
        
        for routine in (self._resume,):
            close_codes_name =  f'_clode_codes_{routine.__name__}'
            close_codes = getattr(self, close_codes_name)
            if not close_code in close_codes:
                continue
            break
        else:
            raise _errors.Interrupted()
        
        await routine()

    async def _listen(self):

        async for message in self._websocket:
            try:
                handle = getattr(self, f'_listen_{message.type.name}')
            except AttributeError:
                continue
            await handle(message.data)

    async def _process(self):

        while not self._websocket.closed:
            await self._listen()
            await self._egress()

    async def _restart(self):

        await self._stop()
        await self._start()

    async def _disconnect(self):

        self._event_identify.clear()
        self._event_complete.clear()

        await self._websocket.close(code = 1000)

    async def _connect(self, uri):

        query = {
            'v': self._version,
            'encoding': self._encoding,
            'compress': 'zlib-stream'
        }

        _helpers.clean_query(query)

        self._websocket = await self._session.ws_connect(uri,  params = query)

    async def _start(self):

        await self._connect(self._uri)

        await self._listen()

    def start(self) -> typing.Awaitable[None]:

        """
        Start the :ddoc:`connection lifecycle </topics/gateway#connection-lifecycle>`.
        """

        return self._start()

    async def _stop(self):

        await self._vital.stop()

        await self._disconnect()

    def stop(self) -> typing.Awaitable[None]:

        """
        Cancel tasks and close connections.
        """

        return self._stop()


import typing

from . import types as _types
from . import enums as _enums


__all__ = (
    'ApplicationCommand', 'ApplicationCommandOption', 
    'ApplicationCommandOptionChoice', 'GuildApplicationCommandPermissions',
    'ApplicationCommandPermission', 'MessageActionRowComponent',
    'MessageButtonComponent', 'MessageSelectMenuComponent', 
    'MessageSelectMenuComponentOption', 'MessageTextInputComponent', 
    'Interaction', 'ApplicationCommandInteractionData', 
    'ApplicationCommandInteractionDataOption', 
    'MessageComponentInteractionData', 'ModalSubmitInteractionData', 
    'ResolvedInteractionData', 'MessageInteraction', 'InteractionResponse', 
    'MessageInteractionResponse', 'AutocompleteInteractionResponse',
    'ModalInteractionResponse', 'Application', 'ApplicationInstallParams', 
    'ApplicationRoleConnectionMetadata', 'AuditLog', 'AuditLogEntry', 
    'OptionalAuditLogEntryInfo', 'AuditLogChange', 'AutoModerationRule', 
    'AutoModerationTriggerMetadata', 'AutoModerationAction', 
    'AutoModerationActionMetadata', 'Channel', 'Message', 'MessageActivity', 
    'MessageReference', 'Reaction', 'FollowedChannel', 'Overwrite', 
    'ThreadMetadata', 'ThreadMember', 'DefaultReaction', 'ForumTag', 'Embed', 
    'EmbedThumbnail', 'EmbedVideo', 'EmbedImage', 'EmbedProvider', 
    'EmbedAuthor', 'EmbedFooter', 'EmbedField', 'Attachment', 
    'AllowedMentions', 'RoleSubscriptionData', 'Emoji', 'Guild', 
    'GuildWidgetSettings', 'GuildWidget', 'GuildMember', 'Integration', 
    'IntegrationAccount', 'IntegrationApplication', 'Ban', 'WelcomeScreen', 
    'WelcomeScreenChannel', 'GuildOnboarding', 'GuildOnboardingPrompt', 
    'GuildOnboardingPromptOption', 'GuildScheduledEvent', 
    'GuildScheduledEventEntityMetadata', 'GuildScheduledEventUser', 
    'GuildTemplate', 'Invite', 'InviteStageInstance', 'StageInstance', 
    'Sticker', 'StickerPack', 'User', 'Connection', 
    'ApplicationRoleConnection', 'VoiceState', 'VoiceRegion', 
    'Webhook', 'Presence', 'ClientStatus', 'Activity', 'ActivityTimestamps', 
    'ActivityParty', 'ActivityAssets', 'ActivitySecrets', 'ActivityButton', 
    'Role', 'RoleTags', 'Team', 'TeamMember'
)


class ApplicationCommand(typing.TypedDict):
    
    id                        : _types.Snowflake
    type                      : _enums.ApplicationCommandType
    application_id            : _types.Snowflake
    guild_id                  : _types.Snowflake
    name                      : _types.String
    name_localizations        : dict[_enums.Locale, _types.String]
    description               : _types.String
    description_localizations : dict[_enums.Locale, _types.String]
    options                   : list['ApplicationCommandOption']
    default_member_permissions: _enums.Permissions
    dm_permission             : _types.Boolean
    default_permission        : _types.Boolean
    nsfw                      : _types.Boolean
    version                   : _types.Snowflake


class ApplicationCommandOption(typing.TypedDict):
    
    type                     : _enums.ApplicationCommandOptionType
    name                     : _types.String
    name_localizations       : dict[_enums.Locale, _types.String]
    description              : _types.String
    description_localizations: dict[_enums.Locale, _types.String]
    required                 : _types.Boolean
    choices                  : list['ApplicationCommandOptionChoice']
    options                  : list['ApplicationCommandOption']
    channel_types            : list[_enums.ChannelType]
    min_value                : typing.Union[_types.Integer, _types.Decimal]
    max_value                : typing.Union[_types.Integer, _types.Decimal]
    min_length               : _types.Integer
    max_length               : _types.Integer
    autocomplete             : _types.Boolean


class ApplicationCommandOptionChoice(typing.TypedDict):
    
    name              : _types.String
    name_localizations: dict[_enums.Locale, _types.String]
    value             : typing.Union[_types.String, _types.Integer, _types.Decimal]


class GuildApplicationCommandPermissions(typing.TypedDict):

    id            : _types.Snowflake
    application_id: _types.Snowflake
    guild_id      : _types.Snowflake
    permissions   : _types.Collection['ApplicationCommandPermission']


class ApplicationCommandPermission(typing.TypedDict):

    id        : _types.Snowflake
    type      : _enums.ApplicationCommandPermissionType
    permission: _types.Boolean


class MessageActionRowComponent(typing.TypedDict):
    
    type      : _enums.MessageComponentType
    components: list[typing.Union['MessageButtonComponent', 'MessageSelectMenuComponent', 'MessageTextInputComponent']]


class MessageButtonComponent(typing.TypedDict):
    
    type     : _types.Integer
    style    : _types.Integer
    label    : _types.String
    emoji    : 'Emoji'
    custom_id: _types.String
    url      : _types.String
    disabled : _types.Boolean


class MessageSelectMenuComponent(typing.TypedDict):
    
    type         : _enums.MessageComponentType
    custom_id    : _types.String
    options      : list['MessageSelectMenuComponentOption']
    channel_types: list[_enums.ChannelType]
    placeholder  : _types.String
    min_values   : _types.Integer
    max_values   : _types.Integer
    disabled     : _types.Boolean


class MessageSelectMenuComponentOption(typing.TypedDict):
    
    label      : _types.String
    value      : _types.String
    description: _types.String
    emoji      : 'Emoji'
    default    : _types.Boolean


class MessageTextInputComponent(typing.TypedDict):
    
    type       : _types.Integer
    custom_id  : _types.String
    style      : _enums.MessageTextInputComponentStyle
    label      : _types.String
    min_length : _types.Integer
    max_length : _types.Integer
    required   : _types.Boolean
    value      : _types.String
    placeholder: _types.String


class Interaction(typing.TypedDict):
    
    id             : _types.Snowflake
    application_id : _types.Snowflake
    type           : _enums.InteractionType
    data           : typing.Union['ApplicationCommandInteractionData', 'MessageComponentInteractionData', 'ModalSubmitInteractionData', 'ResolvedInteractionData']
    guild_id       : _types.Snowflake
    channel        : 'Channel'
    channel_id     : _types.Snowflake
    member         : 'GuildMember'
    user           : 'User'
    token          : _types.String
    version        : _types.Integer
    message        : 'Message'
    app_permissions: _types.String
    locale         : _types.String
    guild_locale   : _types.String


class ApplicationCommandInteractionData(typing.TypedDict):
    
    id       : _types.Snowflake
    name     : _types.String
    type     : _types.Integer
    resolved : 'ResolvedInteractionData'
    options  : list['ApplicationCommandInteractionDataOption']
    guild_id : _types.Snowflake
    target_id: _types.Snowflake


class ApplicationCommandInteractionDataOption(typing.TypedDict):
    
    name   : _types.String
    type   : _types.Integer
    value  : typing.Union[_types.String, _types.Integer, _types.Decimal, _types.Boolean]
    options: list['ApplicationCommandInteractionDataOption']
    focused: _types.Boolean


class MessageComponentInteractionData(typing.TypedDict):
    
    custom_id     : _types.String
    component_type: _types.Integer
    values        : list['MessageSelectMenuComponent']


class ModalSubmitInteractionData(typing.TypedDict):
    
    custom_id : _types.String
    components: list[typing.Union['MessageActionRowComponent', 'MessageButtonComponent', 'MessageSelectMenuComponent', 'MessageTextInputComponent']]


class ResolvedInteractionData(typing.TypedDict):
    
    users        : list['User']
    guild_members: list['GuildMember']
    channels     : list['Channel']
    messages     : list['Message']
    attachments  : list['Attachment']


class MessageInteraction(typing.TypedDict):
    
    id          : _types.Snowflake
    type        : _enums.InteractionType
    name        : _types.String
    user        : 'User'
    guild_member: 'GuildMember'


class InteractionResponse(typing.TypedDict):
    
    type: _enums.InteractionResponseType
    data: typing.Union['MessageInteractionResponse', 'AutocompleteInteractionResponse', 'ModalInteractionResponse']


class MessageInteractionResponse(typing.TypedDict):

    tts             : typing.Optional[_types.Boolean]
    content         : typing.Optional[_types.String]
    embeds          : typing.Optional[list['Embed']]
    allowed_mentions: typing.Optional['AllowedMentions']
    flags           : typing.Optional[_enums.MessageFlags]
    components      : typing.Optional[list[typing.Union['MessageButtonComponent', 'MessageSelectMenuComponent', 'MessageTextInputComponent']]]
    attachments     : typing.Optional[list['Attachment']]


class AutocompleteInteractionResponse(typing.TypedDict):

    choices: list['ApplicationCommandOptionChoice']


class ModalInteractionResponse(typing.TypedDict):

    custom_id : _types.String
    title     : _types.String
    components: list[typing.Union['MessageButtonComponent', 'MessageSelectMenuComponent', 'MessageTextInputComponent']]


class Application(typing.TypedDict):
    
    id                               : _types.Snowflake
    name                             : _types.String
    icon                             : typing.Optional[_types.String]
    description                      : _types.String
    rpc_origins                      : list[_types.String]
    bot_public                       : _types.Boolean
    bot_require_code_grant           : _types.Boolean
    terms_of_service_url             : _types.String
    privacy_policy_url               : _types.String
    owner                            : 'User'
    verify_key                       : _types.String
    team                             : typing.Optional['Team']
    guild_id                         : _types.Snowflake
    primary_sku_id                   : _types.Snowflake
    slug                             : _types.String
    cover_image                      : _types.String
    flags                            : _types.Integer
    tags                             : list[_types.String]
    install_params                   : 'ApplicationInstallParams'
    custom_install_url               : _types.String
    role_connections_verification_url: _types.String


class ApplicationInstallParams(typing.TypedDict):
    
    scopes     : list[_types.String]
    permissions: _enums.Permissions


class ApplicationRoleConnectionMetadata(typing.TypedDict):
    
    type                     : _enums.ApplicationRoleConnectionMetadataType
    key                      : _types.String
    name                     : _types.String
    name_localizations       : dict[_enums.Locale, _types.String]
    description              : _types.String
    description_localizations: dict[_enums.Locale, _types.String]


class AuditLog(typing.TypedDict):
    
    application_commands  : list['ApplicationCommand']
    audit_log_entries     : list['AuditLogEntry']
    auto_moderation_rules : list['AutoModerationRule']
    guild_scheduled_events: list['GuildScheduledEvent']
    integrations          : list['Integration']
    threads               : list['Channel']
    users                 : list['User']
    webhooks              : list['Webhook']


class AuditLogEntry(typing.TypedDict):
    
    target_id  : typing.Optional[_types.String]
    changes    : list['AuditLogChange']
    user_id    : typing.Optional[_types.Snowflake]
    id         : _types.Snowflake
    action_type: _enums.AuditLogEvent
    options    : 'OptionalAuditLogEntryInfo'
    reason     : _types.String


class OptionalAuditLogEntryInfo(typing.TypedDict):
    
    application_id                   : _types.Snowflake
    auto_moderation_rule_name        : _types.String
    auto_moderation_rule_trigger_type: _types.String
    channel_id                       : _types.Snowflake
    count                            : _types.String
    delete_member_days               : _types.String
    id                               : _types.Snowflake
    members_removed                  : _types.String
    message_id                       : _types.Snowflake
    role_name                        : _types.String
    type                             : _enums.OptionalAuditLogEntryInfoOverwrittenEntityType


class AuditLogChange(typing.TypedDict):
    
    new_value: typing.Any
    old_value: typing.Any
    key      : _types.String


class AutoModerationRule(typing.TypedDict):
    
    id              : _types.Snowflake
    guild_id        : _types.Snowflake
    name            : _types.String
    creator_id      : _types.Snowflake
    event_type      : _enums.AutoModerationRuleEventType
    trigger_type    : _enums.AutoModerationTriggerType
    trigger_metadata: 'AutoModerationTriggerMetadata'
    actions         : list['AutoModerationAction']
    enabled         : _types.Boolean
    exempt_roles    : list[_types.Snowflake]
    exempt_channels : list[_types.Snowflake]


class AutoModerationTriggerMetadata(typing.TypedDict):
    
    keyword_filter                 : list[_types.String]
    regex_patterns                 : list[_types.String]
    presets                        : list[_enums.AutoModerationRuleKeywordPresetType]
    allow_list                     : list[_types.String]
    mention_total_limit            : _types.Integer
    mention_raid_protection_enabled: _types.Boolean


class AutoModerationAction(typing.TypedDict):
    
    type    : _enums.AutoModerationActionType
    metadata: 'AutoModerationActionMetadata'


class AutoModerationActionMetadata(typing.TypedDict):
    
    channel_id      : _types.Snowflake
    duration_seconds: _types.Integer
    custom_message  : _types.String


class Channel(typing.TypedDict):
    
    id                                : _types.Snowflake
    type                              : _enums.ChannelType
    guild_id                          : _types.Snowflake
    position                          : _types.Integer
    permission_overwrites             : list['Overwrite']
    name                              : typing.Optional[_types.String]
    topic                             : typing.Optional[_types.String]
    nsfw                              : _types.Boolean
    last_message_id                   : typing.Optional[_types.Snowflake]
    bitrate                           : _types.Integer
    user_limit                        : _types.Integer
    rate_limit_per_user               : _types.Integer
    recipients                        : list['User']
    icon                              : typing.Optional[_types.String]
    owner_id                          : _types.Snowflake
    application_id                    : _types.Snowflake
    managed                           : _types.Boolean
    parent_id                         : typing.Optional[_types.Snowflake]
    last_pin_timestamp                : typing.Optional[_types.ISO8601Timestamp]
    rtc_region                        : typing.Optional[_types.String]
    video_quality_mode                : _enums.ChannelVideoQualityMode
    message_count                     : _types.Integer
    member_count                      : _types.Integer
    thread_metadata                   : 'ThreadMetadata'
    member                            : 'ThreadMember'
    default_auto_archive_duration     : _types.Integer
    permissions                       : _enums.Permissions
    flags                             : _enums.ChannelFlags
    total_message_sent                : _types.Integer
    available_tags                    : list['ForumTag']
    applied_tags                      : list[_types.Snowflake]
    default_reaction_emoji            : typing.Optional['DefaultReaction']
    default_thread_rate_limit_per_user: _types.Integer
    default_sort_order                : typing.Optional[_enums.ChannelSortOrderType]
    default_forum_layout              : _enums.ForumLayoutType


class Message(typing.TypedDict):
    
    id                    : _types.Snowflake
    channel_id            : _types.Snowflake
    author                : 'User'
    content               : _types.String
    timestamp             : _types.ISO8601Timestamp
    edited_timestamp      : typing.Optional[_types.ISO8601Timestamp]
    tts                   : _types.Boolean
    mention_everyone      : _types.Boolean
    mentions              : list['User']
    mention_roles         : list[_types.Snowflake]
    mention_channels      : list['Channel']
    attachments           : list['Attachment']
    embeds                : list['Embed']
    reactions             : list['Reaction']
    nonce                 : _types.String | _types.Integer
    pinned                : _types.Boolean
    webhook_id            : _types.Snowflake
    type                  : _enums.MessageType
    activity              : 'MessageActivity'
    application           : 'Application'
    application_id        : _types.Snowflake
    message_reference     : 'MessageReference'
    flags                 : _enums.MessageFlags
    referenced_message    : typing.Optional['Message']
    interaction           : 'MessageInteraction'
    thread                : 'Channel'
    components            : list['MessageActionRowComponent']
    stickers              : list['Sticker']
    position              : _types.Integer
    role_subscription_data: 'RoleSubscriptionData'


class MessageActivity(typing.TypedDict):
    
    type    : _enums.MessageActivityType
    party_id: _types.String


class MessageReference(typing.TypedDict):
    
    message_id        : _types.Snowflake
    channel_id        : _types.Snowflake
    guild_id          : _types.Snowflake
    fail_if_not_exists: _types.Boolean


class Reaction(typing.TypedDict):
    
    count: _types.Integer
    me   : _types.Boolean
    emoji: 'Emoji'


class FollowedChannel(typing.TypedDict):
    
    channel_id: _types.Snowflake
    webhook_id: _enums.OverwriteType


class Overwrite(typing.TypedDict):
    
    id   : _types.Snowflake
    type : _enums.OverwriteType
    allow: _enums.Permissions
    deny : _enums.Permissions


class ThreadMetadata(typing.TypedDict):
    
    archived             : _types.Boolean
    auto_archive_duration: _types.Integer
    archive_timestamp    : _types.ISO8601Timestamp
    locked               : _types.Boolean
    invitable            : _types.Boolean
    create_timestamp     : typing.Optional[_types.ISO8601Timestamp]


class ThreadMember(typing.TypedDict):
    
    thread_id     : _types.Snowflake
    user_id       : _types.Snowflake
    join_timestamp: _types.ISO8601Timestamp
    flags         : _types.Integer
    member        : 'GuildMember'


class DefaultReaction(typing.TypedDict):
    
    emoji_id  : typing.Optional[_types.Snowflake]
    emoji_name: typing.Optional[_types.String]


class ForumTag(typing.TypedDict):
    
    id        : _types.Snowflake
    name      : _types.String
    moderated : _types.Boolean
    emoji_id  : typing.Optional[_types.Snowflake]
    emoji_name: typing.Optional[_types.String]


class Embed(typing.TypedDict):
    
    title      : _types.String
    type       : _enums.EmbedType
    description: _types.String
    url        : _types.String
    timestamp  : _types.ISO8601Timestamp
    color      : _types.Integer
    footer     : 'EmbedFooter'
    image      : 'EmbedImage'
    thumbnail  : 'EmbedThumbnail'
    video      : 'EmbedVideo'
    provider   : 'EmbedProvider'
    author     : 'EmbedAuthor'
    fields     : list['EmbedField']


class EmbedThumbnail(typing.TypedDict):
    
    url      : _types.String
    proxy_url: _types.String
    height   : _types.Integer
    width    : _types.Integer


class EmbedVideo(typing.TypedDict):
    
    url      : _types.String
    proxy_url: _types.String
    height   : _types.Integer
    width    : _types.Integer


class EmbedImage(typing.TypedDict):
    
    url      : _types.String
    proxy_url: _types.String
    height   : _types.Integer
    width    : _types.Integer


class EmbedProvider(typing.TypedDict):
    
    name: _types.String
    url : _types.String


class EmbedAuthor(typing.TypedDict):
    
    name          : _types.String
    url           : _types.String
    icon_url      : _types.String
    proxy_icon_url: _types.String


class EmbedFooter(typing.TypedDict):
    
    text          : _types.String
    icon_url      : _types.String
    proxy_icon_url: _types.String


class EmbedField(typing.TypedDict):
    
    name  : _types.String
    value : _types.String
    inline: _types.Boolean


class Attachment(typing.TypedDict):
    
    id           : _types.Snowflake
    filename     : _types.String
    description  : _types.String
    content_type : _types.String
    size         : _types.Integer
    url          : _types.String
    proxy_url    : _types.String
    height       : typing.Optional[_types.Integer]
    width        : typing.Optional[_types.Integer]
    ephemeral    : _types.Boolean
    duration_secs: _types.Decimal
    waveform     : _types.String


class AllowedMentions(typing.TypedDict):
    
    parse       : list[_enums.AllowedMentionsType]
    roles       : list[_types.Snowflake]
    users       : list[_types.Snowflake]
    replied_user: _types.Boolean


class RoleSubscriptionData(typing.TypedDict):
    
    role_subscription_listing_id: _types.Snowflake
    tier_name                   : _types.String
    total_months_subscribed     : _types.Integer
    is_renewal                  : _types.Boolean


class Emoji(typing.TypedDict):
    
    id            : typing.Optional[_types.Snowflake]
    name          : typing.Optional[_types.String]
    roles         : list[_types.Snowflake]
    user          : 'User'
    require_colons: _types.Boolean
    managed       : _types.Boolean
    animated      : _types.Boolean
    available     : _types.Boolean


class Guild(typing.TypedDict):
    
    id                           : _types.Snowflake
    name                         : _types.String
    icon                         : typing.Optional[_types.String]
    icon_hash                    : typing.Optional[_types.String]
    splash                       : typing.Optional[_types.String]
    discovery_splash             : typing.Optional[_types.String]
    owner                        : _types.Boolean
    owner_id                     : _types.Snowflake
    permissions                  : _enums.Permissions
    afk_channel_id               : typing.Optional[_types.Snowflake]
    afk_timeout                  : _types.Integer
    widget_enabled               : _types.Boolean
    widget_channel_id            : typing.Optional[_types.Snowflake]
    verification_level           : _enums.GuildVerificationLevel
    default_message_notifications: _enums.GuildDefaultMessageNotificationLevel
    explicit_content_filter      : _enums.GuildExplicitContentFilterLevel
    roles                        : list['Role']
    emojis                       : list['Emoji']
    features                     : list[_enums.GuildFeature]
    mfa_level                    : _types.Integer
    application_id               : typing.Optional[_types.Snowflake]
    system_channel_id            : typing.Optional[_types.Snowflake]
    system_channel_flags         : _enums.GuildSystemChannelFlags
    rules_channel_id             : typing.Optional[_types.Snowflake]
    max_presences                : typing.Optional[_types.Integer]
    max_members                  : _types.Integer
    vanity_url_code              : typing.Optional[_types.String]
    description                  : typing.Optional[_types.String]
    banner                       : typing.Optional[_types.String]
    premium_tier                 : _enums.GuildPremiumTier
    premium_subscription_count   : _types.Integer
    preferred_locale             : _types.String
    public_updates_channel_id    : typing.Optional[_types.Snowflake]
    max_video_channel_users      : _types.Integer
    max_stage_video_channel_users: _types.Integer
    approximate_member_count     : _types.Integer
    approximate_presence_count   : _types.Integer
    welcome_screen               : 'WelcomeScreen'
    nsfw_level                   : _enums.GuildNSFWLevel
    stickers                     : list['Sticker']
    premium_progress_bar_enabled : _types.Boolean
    joined_at                    : _types.ISO8601Timestamp
    large                        : _types.Boolean
    unavailable                  : _types.Boolean
    member_count                 : _types.Integer
    voice_states                 : list['VoiceState']
    members                      : list['GuildMember']
    channels                     : list['Channel']
    threads                      : list['Channel']
    presences                    : list['Presence']
    stage_instances              : list['StageInstance']
    scheduled_events             : list['GuildScheduledEvent']
    safety_alerts_channel_id     : _types.Snowflake


class GuildWidgetSettings(typing.TypedDict):
    
    enabled   : _types.Boolean
    channel_id: typing.Optional[_types.Snowflake]


class GuildWidget(typing.TypedDict):
    
    id            : _types.Snowflake
    name          : _types.String
    instant_invite: typing.Optional[_types.String]
    channels      : list['Channel']
    members       : list['User']
    presence_count: _types.Integer


class GuildMember(typing.TypedDict):
    
    user                        : 'User'
    nick                        : typing.Optional[_types.String]
    avatar                      : typing.Optional[_types.String]
    roles                       : list[_types.Snowflake]
    joined_at                   : _types.ISO8601Timestamp
    premium_since               : typing.Optional[_types.ISO8601Timestamp]
    deaf                        : _types.Boolean
    mute                        : _types.Boolean
    flags                       : _enums.GuildMemberFlags
    pending                     : _types.Boolean
    permissions                 : _enums.Permissions
    communication_disabled_until: typing.Optional[_types.ISO8601Timestamp]


class Integration(typing.TypedDict):
    
    id                 : _types.Snowflake
    name               : _types.String
    type               : _types.String
    enabled            : _types.Boolean
    syncing            : _types.Boolean
    role_id            : _types.Snowflake
    enable_emoticons   : _types.Boolean
    expire_behavior    : _enums.IntegrationExpireBehaviorType
    expire_grace_period: _types.Integer
    user               : 'User'
    account            : 'IntegrationAccount'
    synced_at          : _types.ISO8601Timestamp
    subscriber_count   : _types.Integer
    revoked            : _types.Boolean
    application        : 'Application'
    scopes             : list[_types.String]


class IntegrationAccount(typing.TypedDict):
    
    id  : _types.String
    name: _types.String


class IntegrationApplication(typing.TypedDict):
    
    id         : _types.Snowflake
    name       : _types.String
    icon       : typing.Optional[_types.String]
    description: _types.String
    bot        : 'User'


class Ban(typing.TypedDict):
    
    reason: typing.Optional[_types.String]
    user  : 'User'


class WelcomeScreen(typing.TypedDict):
    
    description     : typing.Optional[_types.String]
    welcome_channels: list['WelcomeScreenChannel']


class WelcomeScreenChannel(typing.TypedDict):
    
    channel_id : _types.Snowflake
    description: _types.String
    emoji_id   : typing.Optional[_types.Snowflake]
    emoji_name : typing.Optional[_types.String]


class GuildOnboarding(typing.TypedDict):
    
    guild_id           : _types.Snowflake
    prompts            : list['GuildOnboardingPrompt']
    default_channel_ids: list[_types.Snowflake]
    enabled            : _types.Boolean


class GuildOnboardingPrompt(typing.TypedDict):
    
    id           : _types.Snowflake
    type         : _enums.GuildOnboardingPromptType
    options      : list['GuildOnboardingPromptOption']
    title        : _types.String
    single_select: _types.Boolean
    required     : _types.Boolean
    in_onboarding: _types.Boolean


class GuildOnboardingPromptOption(typing.TypedDict):
    
    id         : _types.Snowflake
    channel_ids: list[_types.Snowflake]
    role_ids   : list[_types.Snowflake]
    emoji      : 'Emoji'
    title      : _types.String
    description: typing.Optional[_types.String]


class GuildScheduledEvent(typing.TypedDict):
    
    id                  : _types.Snowflake
    guild_id            : _types.Snowflake
    channel_id          : typing.Optional[_types.Snowflake]
    creator_id          : typing.Optional[_types.Snowflake]
    name                : _types.String
    description         : typing.Optional[_types.String]
    scheduled_start_time: _types.ISO8601Timestamp
    scheduled_end_time  : typing.Optional[_types.ISO8601Timestamp]
    privacy_level       : _enums.GuildScheduledEventPrivacyLevel
    status              : _enums.GuildScheduledEventStatus
    entity_type         : _enums.GuildScheduledEventEntityType
    entity_id           : typing.Optional[_types.Snowflake]
    entity_metadata     : typing.Optional['GuildScheduledEventEntityMetadata']
    creator             : 'User'
    user_count          : _types.Integer
    image               : typing.Optional[_types.String]


class GuildScheduledEventEntityMetadata(typing.TypedDict):
    
    location: _types.String


class GuildScheduledEventUser(typing.TypedDict):
    
    guild_scheduled_event_id: _types.Snowflake
    user                    : 'User'
    member                  : 'GuildMember'


class GuildTemplate(typing.TypedDict):
    
    code                   : _types.String
    name                   : _types.String
    description            : typing.Optional[_types.String]
    usage_count            : _types.Integer
    creator_id             : _types.Snowflake
    creator                : 'User'
    created_at             : _types.ISO8601Timestamp
    updated_at             : _types.ISO8601Timestamp
    source_guild_id        : _types.Snowflake
    serialized_source_guild: 'Guild'
    is_dirty               : typing.Optional[_types.Boolean]


class Invite(typing.TypedDict):
    
    code                      : _types.String
    guild                     : 'Guild'
    channel                   : typing.Optional['Channel']
    inviter                   : 'User'
    target_type               : _enums.InviteTargetType
    target_user               : 'User'
    target_application        : 'Application'
    approximate_presence_count: _types.Integer
    approximate_member_count  : _types.Integer
    expires_at                : typing.Optional[_types.ISO8601Timestamp]
    stage_instance            : 'InviteStageInstance'
    guild_scheduled_event     : 'GuildScheduledEvent'
    uses                      : _types.Integer
    max_uses                  : _types.Integer
    max_age                   : _types.Integer
    temporary                 : _types.Boolean
    created_at                : _types.ISO8601Timestamp


class InviteStageInstance(typing.TypedDict):
    
    members          : list['GuildMember']
    participant_count: _types.Integer
    speaker_count    : _types.Integer
    topic            : _types.String


class StageInstance(typing.TypedDict):
    
    id                      : _types.Snowflake
    guild_id                : _types.Snowflake
    channel_id              : _types.Snowflake
    topic                   : _types.String
    privacy_level           : _enums.StageInstancePrivacyLevel
    discoverable_disabled   : _types.Boolean
    guild_scheduled_event_id: typing.Optional[_types.Snowflake]


class Sticker(typing.TypedDict):
    
    id         : _types.Snowflake
    pack_id    : _types.Snowflake
    name       : _types.String
    description: typing.Optional[_types.String]
    tags       : _types.String
    asset      : _types.String
    type       : _enums.StickerType
    format_type: _enums.StickerFormatType
    available  : _types.Boolean
    guild_id   : _types.Snowflake
    user       : 'User'
    sort_value : _types.Integer


class StickerPack(typing.TypedDict):
    
    id              : _types.Snowflake
    stickers        : list['Sticker']
    name            : _types.String
    sku_id          : _types.Snowflake
    cover_sticker_id: _types.Snowflake
    description     : _types.String
    banner_asset_id : _types.Snowflake


class User(typing.TypedDict):
    
    id           : _types.Snowflake
    username     : _types.String
    discriminator: _types.String
    avatar       : typing.Optional[_types.String]
    bot          : _types.Boolean
    system       : _types.Boolean
    mfa_enabled  : _types.Boolean
    banner       : typing.Optional[_types.String]
    accent_color : typing.Optional[_types.Integer]
    locale       : _enums.Locale
    verified     : _types.Boolean
    email        : typing.Optional[_types.String]
    flags        : _enums.UserFlags
    premium_type : _enums.UserPremiumType
    public_flags : _enums.UserFlags


class Connection(typing.TypedDict):
    
    id           : _types.String
    name         : _types.String
    type         : _types.String
    revoked      : _types.Boolean
    integrations : list['Integration']
    verified     : _types.Boolean
    friend_sync  : _types.Boolean
    show_activity: _types.Boolean
    two_way_link : _types.Boolean
    visibility   : _enums.ConnectionVisibilityType


class ApplicationRoleConnection(typing.TypedDict):
    
    platform_name    : typing.Optional[_types.String]
    platform_username: typing.Optional[_types.String]
    metadata         : dict[_types.String, 'ApplicationRoleConnectionMetadata']


class VoiceState(typing.TypedDict):
    
    guild_id                  : _types.Snowflake
    channel_id                : typing.Optional[_types.Snowflake]
    user_id                   : _types.Snowflake
    member                    : 'GuildMember'
    session_id                : _types.String
    deaf                      : _types.Boolean
    mute                      : _types.Boolean
    self_deaf                 : _types.Boolean
    self_mute                 : _types.Boolean
    self_stream               : _types.Boolean
    self_video                : _types.Boolean
    suppress                  : _types.Boolean
    request_to_speak_timestamp: typing.Optional[_types.ISO8601Timestamp]


class VoiceRegion(typing.TypedDict):
    
    id        : _types.String
    name      : _types.String
    optimal   : _types.Boolean
    deprecated: _types.Boolean
    custom    : _types.Boolean


class Webhook(typing.TypedDict):
    
    id            : _types.Snowflake
    type          : _enums.WebhookType
    guild_id      : typing.Optional[_types.Snowflake]
    channel_id    : typing.Optional[_types.Snowflake]
    user          : 'User'
    name          : typing.Optional[_types.String]
    avatar        : typing.Optional[_types.String]
    token         : _types.String
    application_id: typing.Optional[_types.Snowflake]
    source_guild  : 'Guild'
    source_channel: 'Channel'
    url           : _types.String


class Presence(typing.TypedDict):
    
    user         : 'User'
    guild_id     : _types.Snowflake
    status       : _enums.StatusType
    activities   : list['Activity']
    client_status: 'ClientStatus'


class ClientStatus(typing.TypedDict):
    
    desktop: _types.String
    mobile : _types.String
    web    : _types.String


class Activity(typing.TypedDict):
    
    name          : _types.String
    type          : _enums.ActivityType
    url           : typing.Optional[_types.String]
    created_at    : _types.Timestamp
    timestamps    : 'ActivityTimestamps'
    application_id: _types.Snowflake
    details       : typing.Optional[_types.String]
    state         : typing.Optional[_types.String]
    emoji         : typing.Optional['Emoji']
    party         : 'ActivityParty'
    assets        : 'ActivityAssets'
    secrets       : 'ActivitySecrets'
    instance      : _types.Boolean
    flags         : _enums.ActivityFlags
    buttons       : list['ActivityButton']


class ActivityTimestamps(typing.TypedDict):
    
    start: _types.Timestamp
    end  : _types.Timestamp


class ActivityParty(typing.TypedDict):
    
    id  : _types.String
    size: list[_types.Integer]


class ActivityAssets(typing.TypedDict):
    
    large_image: _types.String
    large_text : _types.String
    small_image: _types.String
    small_text : _types.String


class ActivitySecrets(typing.TypedDict):
    
    join    : _types.String
    spectate: _types.String
    match   : _types.String


class ActivityButton(typing.TypedDict):
    
    label: _types.String
    url  : _types.String


class Role(typing.TypedDict):
    
    id           : _types.Snowflake
    name         : _types.String
    color        : _types.Integer
    hoist        : _types.Boolean
    icon         : typing.Optional[_types.String]
    unicode_emoji: typing.Optional[_types.String]
    position     : _types.Integer
    permissions  : _types.String
    managed      : _types.Boolean
    mentionable  : _types.Boolean
    tags         : 'RoleTags'


class RoleTags(typing.TypedDict):
    
    bot_id                 : _types.Snowflake
    integration_id         : _types.Snowflake
    premium_subscriber     : _types.Boolean
    subscription_listing_id: _types.Snowflake
    available_for_purchase : _types.Boolean
    guild_connections      : _types.Boolean


class Team(typing.TypedDict):
    
    icon         : typing.Optional[_types.String]
    id           : _types.Snowflake
    members      : list['TeamMember']
    name         : _types.String
    owner_user_id: _types.Snowflake


class TeamMember(typing.TypedDict):
    
    membership_state: _enums.TeamMemberMembershipState
    permissions     : list[_types.String]
    team_id         : _types.Snowflake
    user            : 'User'


import aiocord
import secrets

@aiocord.widget.interact('shop')
async def _interact_shop(info: aiocord.widget.Info, core_event: aiocord.events.CreateInteraction):
    # define the stock
    stock = {
        'ice-cream': {
            'name': 'Ice Cream',
            'emoji': '🍦'
        },
        'cake': {
            'name': 'Cake',
            'emoji': '🍰'
        }, 
        'waffle': {
            'name': 'Waffle',
            'emoji': '🧇'
        }
    }
    # unique identifier
    select_custom_id = secrets.token_hex(16)
    # compose the response
    response = aiocord.model.protocols.InteractionResponse(
        type = aiocord.model.enums.InteractionResponseType.channel_message_with_source,
        data = aiocord.model.protocols.MessageInteractionResponse(
            content = 'What would you like to buy?',
            components = [
                aiocord.model.protocols.MessageActionRowComponent(
                    type = aiocord.model.enums.MessageComponentType.action_row,
                    components = [
                        aiocord.model.protocols.MessageSelectMenuComponent(
                            custom_id = select_custom_id,
                            type = aiocord.model.enums.MessageComponentType.string_select,
                            options = [
                                aiocord.model.protocols.MessageSelectMenuComponentOption(
                                    value = key,
                                    label = item['name'],
                                    description = 'Sweet, fluffy, chunky and super tasty.',
                                    emoji = aiocord.model.protocols.Emoji(
                                        name = item['emoji']
                                    )
                                )
                                for key, item in stock.items()
                            ]
                        )
                    ]
                )
            ],
            # ensure only the invoker can see this
            flags = aiocord.model.enums.MessageFlags.ephemeral
        )
    )
    # condition for the expected interaction
    async def check(core_event: aiocord.events.CreateInteraction):
        # only accept an interaction to the identifer
        return core_event.interaction.type == aiocord.model.enums.InteractionType.message_component and core_event.interaction.data.custom_id == select_custom_id
    # make the waiter
    sentinel = info.client.wait(aiocord.events.CreateInteraction, check)
    # send the response
    await info.client.create_interaction_response(core_event.interaction.id, core_event.interaction.token, **response)
    # wait the interaction
    core_event = await sentinel
    # resolve the picked item
    item = stock[core_event.interaction.data.values[0]]['name']
    # unique identifiers
    negative_button_custom_id = secrets.token_hex(16)
    positive_button_custom_id = secrets.token_hex(16)
    # compose the response
    response = aiocord.model.protocols.InteractionResponse(
        type = aiocord.model.enums.InteractionResponseType.channel_message_with_source,
        data = aiocord.model.protocols.MessageInteractionResponse(
            content = 'Is it a gift for someone?',
            components = [
                aiocord.model.protocols.MessageActionRowComponent(
                    type = aiocord.model.enums.MessageComponentType.action_row,
                    components = [
                        aiocord.model.protocols.MessageButtonComponent(
                            custom_id = positive_button_custom_id,
                            type = aiocord.model.enums.MessageComponentType.button,
                            label = 'Yes',
                            style = aiocord.model.enums.MessageButtonComponentStyle.primary
                        ),
                        aiocord.model.protocols.MessageButtonComponent(
                            custom_id = negative_button_custom_id,
                            type = aiocord.model.enums.MessageComponentType.button,
                            label = 'No',
                            style = aiocord.model.enums.MessageButtonComponentStyle.secondary
                        )
                    ]
                )
            ],
            # ensure only the invoker can see this
            flags = aiocord.model.enums.MessageFlags.ephemeral
        )
    )
    # unique identifier group
    custom_id_button_group = [negative_button_custom_id, positive_button_custom_id]
    # condition for the expected interaction
    async def check(core_event: aiocord.events.CreateInteraction):
        # only accept an interaction to the identifers
        return core_event.interaction.type == aiocord.model.enums.InteractionType.message_component and core_event.interaction.data.custom_id in custom_id_button_group
    # make the waiter
    sentinel = info.client.wait(aiocord.events.CreateInteraction, check)
    # send the response
    await info.client.create_interaction_response(core_event.interaction.id, core_event.interaction.token, **response)
    # wait the interaction
    core_event = await sentinel
    # resolve whether positive
    is_gift = custom_id_button_group.index(core_event.interaction.data.custom_id)
    # ...is it?
    if is_gift:
        # unique identifier
        select_custom_id = secrets.token_hex(16)
        # compose the response
        response = aiocord.model.protocols.InteractionResponse(
            type = aiocord.model.enums.InteractionResponseType.channel_message_with_source,
            data = aiocord.model.protocols.MessageInteractionResponse(
                content = 'Who is this a gift for?',
                components = [
                    aiocord.model.protocols.MessageActionRowComponent(
                        type = aiocord.model.enums.MessageComponentType.action_row,
                        components = [
                            aiocord.model.protocols.MessageSelectMenuComponent(
                                custom_id = select_custom_id,
                                type = aiocord.model.enums.MessageComponentType.user_select
                            )
                        ]
                    )
                ],
                flags = aiocord.model.enums.MessageFlags.ephemeral
            )
        )
        # condition for the expected interaction
        async def check(core_event: aiocord.events.CreateInteraction):
            return core_event.interaction.type == aiocord.model.enums.InteractionType.message_component and core_event.interaction.data.custom_id == select_custom_id
        # make the waiter
        sentinel = info.client.wait(aiocord.events.CreateInteraction, check)
        # send the response
        await info.client.create_interaction_response(core_event.interaction.id, core_event.interaction.token, **response)
        # wait the interaction
        core_event = await sentinel
        # resolve the recipients
        recipient = ' '.join(aiocord.model.mentions.user(user_id) for user_id in core_event.interaction.data.values)
    else:
        recipient = 'themselves'
    # compose the response
    response = aiocord.model.protocols.InteractionResponse(
        type = aiocord.model.enums.InteractionResponseType.channel_message_with_source,
        data = aiocord.model.protocols.MessageInteractionResponse(
            # anyone can see this
            content = f'{info.client.cache.user.mention()} bought {item} for {recipient}!'
        )
    )
    # send the response
    await info.client.create_interaction_response(core_event.interaction.id, core_event.interaction.token, **response)


import aiocord
import functools

async def _interact_shop_respond(item, recipient, info: aiocord.widget.Info, core_event: aiocord.events.CreateInteraction):

    response = aiocord.model.protocols.InteractionResponse(
        type = aiocord.model.enums.InteractionResponseType.channel_message_with_source,
        data = aiocord.model.protocols.MessageInteractionResponse(
            content = f'{core_event.interaction.member.user.mention()} bought {item} for {recipient}!'
        )
    )

    return response

async def _interact_shop_item_gifting_select(item, info: aiocord.widget.Info, core_event: aiocord.events.CreateInteraction):
    # ignore timeouts
    if core_event is None:
        return
    # compose the recipient
    recipient = ' '.join(aiocord.model.mentions.user(user_id) for user_id in core_event.interaction.data.values)
    # get the response
    response = await _interact_shop_respond(item, recipient, info, core_event)
    # send the response
    return response

async def _interact_shop_item_gifting_accept(item, info: aiocord.widget.Info, core_event: aiocord.events.CreateInteraction):
    # ignore timeouts
    if core_event is None:
        return
    # compose the response
    response = aiocord.model.protocols.InteractionResponse(
        type = aiocord.model.enums.InteractionResponseType.channel_message_with_source,
        data = aiocord.model.protocols.MessageInteractionResponse(
            content = 'Who is this a gift for?',
            components = [
                aiocord.model.protocols.MessageActionRowComponent(
                    type = aiocord.model.enums.MessageComponentType.action_row,
                    components = [
                        aiocord.utils.interact(
                            info.client,
                            functools.partial(_interact_shop_item_gifting_select, item, info),
                            aiocord.model.protocols.MessageSelectMenuComponent(
                                type = aiocord.model.enums.MessageComponentType.user_select
                            )
                        )
                    ]
                )
            ],
            flags = aiocord.model.enums.MessageFlags.ephemeral
        )
    )
    # send the response
    return response

async def _interact_shop_item_gifting_reject(item, info: aiocord.widget.Info, core_event: aiocord.events.CreateInteraction):
    # ignore timeouts
    if core_event is None:
        return
    # compose the recipient
    recipient = 'themselves'
    # get the response
    response = await _interact_shop_respond(item, recipient, info, core_event)
    # send the response
    return response

async def _interact_shop_stock_select(stock, info: aiocord.widget.Info, core_event: aiocord.events.CreateInteraction):
    # ignore timeouts
    if core_event is None:
        return
    # get the item name from stock
    item = stock[core_event.interaction.data.values[0]]['name']
    # compose the response
    response = aiocord.model.protocols.InteractionResponse(
        type = aiocord.model.enums.InteractionResponseType.channel_message_with_source,
        data = aiocord.model.protocols.MessageInteractionResponse(
            content = 'Is it a gift for someone?',
            components = [
                aiocord.model.protocols.MessageActionRowComponent(
                    type = aiocord.model.enums.MessageComponentType.action_row,
                    components = [
                        aiocord.utils.interact(
                            info.client,
                            functools.partial(_interact_shop_item_gifting_accept, item, info),
                            aiocord.model.protocols.MessageButtonComponent(
                                type = aiocord.model.enums.MessageComponentType.button,
                                label = 'Yes',
                                style = aiocord.model.enums.MessageButtonComponentStyle.primary
                            )
                        ),
                        aiocord.utils.interact(
                            info.client,
                            functools.partial(_interact_shop_item_gifting_reject, item, info),
                            aiocord.model.protocols.MessageButtonComponent(
                                type = aiocord.model.enums.MessageComponentType.button,
                                label = 'No',
                                style = aiocord.model.enums.MessageButtonComponentStyle.secondary
                            )
                        ),
                    ]
                )
            ],
            flags = aiocord.model.enums.MessageFlags.ephemeral
        )
    )
    # send the response
    return response

@aiocord.widget.interact('shop')
async def _interact_shop(info: aiocord.widget.Info, core_event: aiocord.events.CreateInteraction):
    # define the stock
    stock = {
        'ice-cream': {
            'name': 'Ice Cream',
            'emoji': '🍦'
        },
        'cake': {
            'name': 'Cake',
            'emoji': '🍰'
        }, 
        'waffle': {
            'name': 'Waffle',
            'emoji': '🧇'
        }, 
    }
    # compose the response
    response = aiocord.model.protocols.InteractionResponse(
        type = aiocord.model.enums.InteractionResponseType.channel_message_with_source,
        data = aiocord.model.protocols.MessageInteractionResponse(
            content = 'What would you like to buy?',
            components = [
                aiocord.model.protocols.MessageActionRowComponent(
                    type = aiocord.model.enums.MessageComponentType.action_row,
                    components = [
                        aiocord.utils.interact(
                            info.client,
                            functools.partial(_interact_shop_stock_select, stock, info),
                            aiocord.model.protocols.MessageSelectMenuComponent(
                                type = aiocord.model.enums.MessageComponentType.string_select,
                                options = [
                                    aiocord.model.protocols.MessageSelectMenuComponentOption(
                                        value = key,
                                        label = item['name'],
                                        description = 'Sweet, fluffy, chunky and super tasty.',
                                        emoji = aiocord.model.protocols.Emoji(
                                            name = item['emoji']
                                        )
                                    )
                                    for key, item in stock.items()
                                ]
                            )
                        )
                    ]
                )
            ],
            flags = aiocord.model.enums.MessageFlags.ephemeral
        )
    )
    # send the response
    return response


import typing
import asyncio
import itertools


__all__ = ('Vital',)


class Vital:

    __slots__ = ('_beat', '_save', '_task', '_dead')

    def __init__(self, beat, save):

        self._beat = beat
        self._save = save

        self._task = None
        self._dead = False

    @property
    def dead(self):

        return self._dead
    
    def _ack(self):

        self._dead = False 

    def ack(self):

        self._ack()

    async def _pulse(self, jitter, interval):

        for cycle in itertools.count(0):
            jit = jitter(cycle) if jitter else 1
            await asyncio.sleep(interval * jit)
            if self._dead:
                await self._save()
            else:
                self._dead = True
                await self._beat()

    def _start(self, jitter, interval):

        coro = self._pulse(jitter, interval)

        self._task = asyncio.create_task(coro)

    def start(self,
              interval: int,
              jitter  : typing.Callable[[int], float] = None) -> None:

        return self._start(jitter, interval)

    async def _stop(self):

        if self._task is None:
            return

        self._task.cancel()

        try:
            await self._task
        except asyncio.CancelledError:
            pass

        self._task = None

    def stop(self) -> typing.Awaitable[None]:

        return self._stop()




__all__ = (
    'Base', 'Client', 'Connectipn', 'Interrupted', 'Invalidated'
)


class Base(Exception):

    """
    Base module error.
    """

    __slots__ = ()


class Client(Base):

    """
    Base for client errors.
    """

    __slots__ = ()


class Connection(Client):

    """
    Base for connection-related errors.
    """

    __slots__ = ()


class Interrupted(Connection):

    """
    Received a non-salvageable :ddoc:`close code </topics/opcodes-and-status-codes#voice-voice-close-event-codes>`.
    """

    __slots__ = ()


class Invalidated(Connection):

    """
    Received a non-salvageable :ddoc:`invalid session </topics/gateway-events#invalid-session>`.
    """

    __slots__ = ()